from __future__ import division

import numpy as np
from ..constants import molwt

def regress(emissions,
            beta=np.array([2.8249e-4, 1.0695e-4, -9.3604e-4, 99.7831e-4])):

    """Calculates tropospheric ozone forcing from precursor emissions.

    Inputs: (nt x 40) emissions array

    Keywords:
        beta: 4-element array of regression coefficients of precursor
              radiative efficiency, W m-2 (Mt yr-1)-1.
              order is [CH4, CO, NMVOC, NOx]

    Outputs:
        tropospheric ozone ERF time series.
    """

    if emissions.ndim==2:
        em_CH4, em_CO, em_NMVOC, em_NOx = emissions[:,[3, 6, 7, 8]].T
    else:
        em_CH4, em_CO, em_NMVOC, em_NOx = emissions[[3, 6, 7, 8]]

    F_CH4   = beta[0] * em_CH4
    F_CO    = beta[1] * em_CO
    F_NMVOC = beta[2] * em_NMVOC
    F_NOx   = beta[3] * em_NOx

    F = F_CH4 + F_CO + F_NMVOC + F_NOx
    return F


def cmip6_stevenson(emissions, C_CH4, T=0, feedback=False,
    PI=np.array([722, 170, 10, 4.29]), 
    beta=np.array([1.77871043e-04, 5.80173377e-05, 2.09151270e-03,
        1.94458719e-04])):
    """Calculates tropospheric ozone forcing from precursor emissions based on
    Stevenson et al, 2013 10.5194/acp-13-3063-2013

    Inputs:
        emissions: (nt x 40) numpy array     
        C_CH4    : (nt) numpy array of methane concentrations, ppb

    Keywords:
        T              : change in surface temperature since pre-industrial
        feedback       : True or False - include temperature feedback on ozone
                         forcing?
        PI:            : 4-element array of pre-industrial CH4 concentrations,
                         CO emissions, NMVOC emissions and NOx emissions
        beta:          : coefficients of how CH4 concentrations, CO emissions,
                         NMVOC emissions and NOx emissions affect forcing

    Outputs:
        tropospheric ozone ERF time series.
    """

    # expand to 2D/1D if not already
    if emissions.ndim == 1:
        nspec = len(emissions)
        emissions = emissions.reshape((1, nspec))
    if np.isscalar(C_CH4):
        C_CH4 = np.ones(1)*C_CH4

    year, em_CO, em_NMVOC, em_NOx = emissions[:,[0, 6, 7, 8]].T
    nt = len(year)
    F_CH4, F_CO, F_NMVOC, F_NOx = np.zeros((4,nt))

    for i in range(nt):
        F_CH4[i]   = beta[0] * (C_CH4[i]-PI[0])
        F_CO[i]    = beta[1] * (em_CO[i]-PI[1])
        F_NMVOC[i] = beta[2] * (em_NMVOC[i]-PI[2])
        F_NOx[i]   = beta[3] * (em_NOx[i]-PI[3])

    # Include the effect of climate feedback? We fit a curve to the 2000, 2030
    # and 2100 best estimates of feedback based on middle-of-the-road
    # temperature projections.
    def temperature_feedback(T, a=0.03189267, b=1.34966941, c=-0.03214807):
        if T<=0:
            return 0
        else:
            return a*np.exp(-b*T)+c

    if feedback:
        F = F_CH4 + F_CO + F_NMVOC + F_NOx + temperature_feedback(T)
    else:
        F = F_CH4 + F_CO + F_NMVOC + F_NOx

    return F


def stevenson(emissions, C_CH4, T=0, feedback=False, fix_pre1850_RCP=False,
    PI=np.array([722, 170, 10, 4.29])):
    """Calculates tropospheric ozone forcing from precursor emissions based on
    Stevenson et al, 2013 10.5194/acp-13-3063-2013

    Inputs:
        emissions: (nt x 40) numpy array     
        C_CH4    : (nt) numpy array of methane concentrations, ppb

    Keywords:
        T              : change in surface temperature since pre-industrial
        feedback       : True or False - include temperature feedback on ozone
                         forcing?
        fix_pre1850_RCP: Use different relationship for 1750/65 to 1850 based 
                         on anthropogenic emissions from Skeie et al (2011)
                         for 1750 (atmos-chem-phys.net/11/11827/2011)
        PI:            : 4-element array of pre-industrial CH4 concentrations,
                         CO emissions, NMVOC emissions and NOx emissions

    Outputs:
        tropospheric ozone ERF time series.
    """

    # expand to 2D/1D if not already
    if emissions.ndim == 1:
        nspec = len(emissions)
        emissions = emissions.reshape((1, nspec))
    if np.isscalar(C_CH4):
        C_CH4 = np.ones(1)*C_CH4

    # numbers in denominator are 2000-1750 concs or emissions used in 
    # Stevenson and traced back to Lamarque et al 2010 for 2000
    # https://www.atmos-chem-phys.net/10/7017/2010/
    year, em_CO, em_NMVOC, em_NOx = emissions[:,[0, 6, 7, 8]].T
    nt = len(year)
    F_CH4, F_CO, F_NMVOC, F_NOx = np.zeros((4,nt))

    for i in range(nt):
        if year[i]>=1850 or fix_pre1850_RCP==False:
            F_CH4[i]   = 0.166/960    * (C_CH4[i]-PI[0])
            F_CO[i]    = 0.058/681.8 * (em_CO[i]-PI[1])
            F_NMVOC[i] = 0.035/155.84 * (em_NMVOC[i]-PI[2])
            F_NOx[i]   = 0.119/61.16  * (em_NOx[i] * 
                molwt.NO / molwt.N - PI[3])
        # The RCP scenarios give a negative forcing prior to ~1780. This is 
        # because the anthropogenic emissions are given to be zero in RCPs but
        # not zero in the Skeie numbers which are used here. This can be fixed
        # to give a more linear behaviour.
        else:
            F_CH4[i]   = 0.166/960    * (C_CH4[i]-722)
            F_CO[i]    = 0.058/681.8 * 215.59  * em_CO[i] / 385.59
            F_NMVOC[i] = 0.035/155.84 * 51.97 * em_NMVOC[i] / 61.97
            F_NOx[i]   = 0.119/61.16  * 7.31 * (em_NOx[i]
                 * molwt.NO / molwt.N) / 11.6

    # Include the effect of climate feedback? We fit a curve to the 2000, 2030
    # and 2100 best estimates of feedback based on middle-of-the-road
    # temperature projections.
    def temperature_feedback(T, a=0.03189267, b=1.34966941, c=-0.03214807):
        if T<=0:
            return 0
        else:
            return a*np.exp(-b*T)+c

    if feedback:
        F = F_CH4 + F_CO + F_NMVOC + F_NOx + temperature_feedback(T)
    else:
        F = F_CH4 + F_CO + F_NMVOC + F_NOx

    return F



from __future__ import division

import numpy as np
from scipy import stats

def hist_temp(Tobs, Tmodel, years, inflate=True, CI=0.9):
    """Checks to see whether model-derived temperatures fall in observational
    uncertainty.

    Uses the method of Thompson et al., 2015, also used in IPCC AR5 to derive
    temperature trends which includes autocorrelation. The regression slope of
    the observations is compared to the regression slope of the model. If the
    trend of the model is within observational uncertainty, the test passes.

    Reference: J. Climate, 28, 6443-6456 10.1175/JCLI-D-14-00830.1

    inputs:
        Tobs: observed temperature time series. Numpy array
        Tmodel: modelled temperature time series. Numpy array
        years: Numpy array of years covered by Tobs and Tmodel

    keywords:
        inflate: True (default) if the uncertainty bounds should be inflated
            for lag-1 autocorrelation - as used in Thompson.
        CI: confidence interval around the mean regression slope to count as
            constrained. Default 0.9.

    returns:
        accept: True if ensemble member agrees with observations else False.
        slope_m: regression slope of modelled temperature
        intercept_m: intercept of modelled temperature
        slope_o: regression slope of observed temperature
        intercept_o: intercept of observed temperature
    """

    n = float(len(years))

    # detrend the 1880-2016 observations and apply the internal variability
    # estimate as detailed in eq. 8 of Thompson et al, J. Climate 2015

    slope_o, intercept_o, _, _, stderr_o = stats.linregress(years,Tobs)
    resid = slope_o * years + intercept_o - Tobs
    s = np.std(resid, ddof=2)
    if inflate:
        lag1ac = np.corrcoef(resid[:-1],resid[1:])[1,0]
    else:
        lag1ac = 0
    gamma = np.sqrt( n / ((n-2) * (1-lag1ac)/(1+lag1ac) - 2 ))
    g = np.sqrt(12./(n**3 - n))
    tcrit = stats.t.ppf(1-(1-CI)/2.0, df=n)
    CI_o = s * gamma * tcrit * g

    # check modelled T against observed
    slope_m,intercept_m,_,_,stderr_m = stats.linregress(years, Tmodel)

    if slope_o-CI_o <= slope_m <= slope_o+CI_o:
        accept = True
    else:
        accept = False
    return accept, slope_m, intercept_m, slope_o, intercept_o



from __future__ import division

import numpy as np

def forcing_to_temperature(t0, q, d, f, e=1.0):
    """Calculate temperature from a given radiative forcing.

    This follows the forcing to temperature function described in Millar et
    al. (2015; 2017).

    Inputs:
        t0: Temperature in timestep t-1
        q: The matrix contributions to slow and fast temperature change
           calculated from ECS and TCR (2 element array)
        d: The slow and fast thermal response time constants (2 element array)
        f: radiative forcing (can be scalar or 1D array representing multiple
           species)

    Keywords:
        e: efficacy factor (default 1); if f is an array, e should be an array
           of the same length.

    Outputs:
        t1: slow and fast contributions to total temperature (2 element array)
        in timestep t
    """
    t1 = t0*np.exp(-1.0/d) + q*(1.0-np.exp((-1.0)/d))*np.sum(f*e)
    return t1


def calculate_q(tcrecs, d, f2x, tcr_dbl, nt):
    """If TCR and ECS are supplied, calculate the q model coefficients.
    See Eqs. (4) and (5) of Millar et al ACP (2017).

    Inputs:
        tcrecs  : 2-element array of transient climate response (TCR) and
                  equilibrium climate sensitivity (ECS).
        d       : The slow and fast thermal response time constants
        f2x     : Effective radiative forcing from a doubling of CO2
        tcr_dbl : time to a doubling of CO2 under 1% per year CO2 increase, yr
        nt      : number of timesteps

    Outputs:
        q       : coefficients of slow and fast temperature change in each
                  timestep ((nt, 2) array).
    """

    # TODO:
    # error checking before call
    # benchmark one call per timestep and if not slower do not convert to 2D
    #  - will make code cleaner

    k = 1.0 - (d/tcr_dbl)*(1.0 - np.exp(-tcr_dbl/d))
    # if ECS and TCR are not time-varying, expand them to 2D array anyway
    if tcrecs.ndim==1:
        if len(tcrecs)!=2:
            raise ValueError(
              "Constant TCR and ECS should be a 2-element array")
        tcrecs = np.ones((nt, 2)) * tcrecs
    elif tcrecs.ndim==2:
        if tcrecs.shape!=(nt, 2):
            raise ValueError(
              "Transient TCR and ECS should be a nt x 2 array")
    q  = (1.0 / f2x) * (1.0/(k[0]-k[1])) * np.array([
        tcrecs[:,0]-tcrecs[:,1]*k[1],tcrecs[:,1]*k[0]-tcrecs[:,0]]).T
    return q




from __future__ import division

import numpy as np

from ..constants import radeff

def etminan(C, Cpi, F2x=3.71, scale_F2x=True):
    """Calculate the radiative forcing from CO2, CH4 and N2O.

    This function uses the updated formulas of Etminan et al. (2016),
    including the overlaps between CO2, methane and nitrous oxide.

    Reference: Etminan et al, 2016, JGR, doi: 10.1002/2016GL071930

    Inputs:
        C: [CO2, CH4, N2O] concentrations, [ppm, ppb, ppb]
        Cpi: pre-industrial [CO2, CH4, N2O] concentrations

    Keywords:
        F2x: radiative forcing from a doubling of CO2.

    Returns:
        3-element array of radiative forcing: [F_CO2, F_CH4, F_N2O]
    """

    Cbar = 0.5 * (C[0] + Cpi[0])
    Mbar = 0.5 * (C[1] + Cpi[1])
    Nbar = 0.5 * (C[2] + Cpi[2])

    # Tune the coefficient of CO2 forcing to acheive desired F2x, using 
    # pre-industrial CO2 and N2O. F2x_etminan ~= 3.801.
    scaleCO2 = 1
    if scale_F2x:
        F2x_etminan = (
          -2.4e-7*Cpi[0]**2 + 7.2e-4*Cpi[0] - 2.1e-4*Cpi[2] + 5.36) * np.log(2)
        scaleCO2 = F2x/F2x_etminan

    F = np.zeros(3)
    F[0] = (-2.4e-7*(C[0] - Cpi[0])**2 + 7.2e-4*np.fabs(C[0]-Cpi[0]) - \
      2.1e-4 * Nbar + 5.36) * np.log(C[0]/Cpi[0]) * scaleCO2
    F[1] = (-1.3e-6*Mbar - 8.2e-6*Nbar + 0.043) * (np.sqrt(C[1]) - \
      np.sqrt(Cpi[1]))
    F[2] = (-8.0e-6*Cbar + 4.2e-6*Nbar - 4.9e-6*Mbar + 0.117) * \
      (np.sqrt(C[2]) - np.sqrt(Cpi[2]))

    return F


def MN(M, N):
    return 0.47 * np.log(1 + 2.01e-5*(M*N)**(0.75) + 5.31e-15*M*(M*N)**(1.52))


def co2_log(C, Cpi, F2x=3.71):
    return F2x/np.log(2) * np.log(C/Cpi)


def myhre(C, Cpi, F2x=3.71, scale_F2x=None):
# TODO: remove scale_F2x in v1.6
    """Calculate the radiative forcing from CO2, CH4 and N2O.

    This uses the Myhre et al. (1998) relationships including the band
    overlaps between CH4 and N2O. It is also used in AR5.

    Reference: Myhre et al, 1998, JGR, doi: 10.1029/98GL01908

    Inputs:
        C: [CO2, CH4, N2O] concentrations, [ppm, ppb, ppb]
        Cpi: pre-industrial [CO2, CH4, N2O] concentrations

    Keywords:
        F2x: radiative forcing from a doubling of CO2.
        scale_F2x: redundant; included for compatibility on import
            in fair_scm

    Returns:
        3-element array of radiative forcing: [F_CO2, F_CH4, F_N2O]
    """

    F = np.zeros(3)

    F[0] = co2_log(C[0], Cpi[0], F2x)
    F[1] = 0.036 * (np.sqrt(C[1]) - np.sqrt(Cpi[1])) - (
      MN(C[1],Cpi[2]) - MN(Cpi[1],Cpi[2]))
    F[2] = 0.12 * (np.sqrt(C[2]) - np.sqrt(Cpi[2])) - (
      MN(Cpi[1],C[2]) - MN(Cpi[1],Cpi[2])) 

    return F


def minor_gases(C, Cpi):
    """
    Calculate radiative forcing from minor gas species.

    Inputs:
        C: concentration of minor GHGs (in order of MAGICC RCP concentration
            spreadsheets)
        Cpi: Pre-industrial concentrations of GHGs
    Returns:
        28 element array of minor GHG forcings
    """

    return (C - Cpi) * radeff.aslist[3:] * 0.001



from __future__ import division

import numpy as np
from ..constants import molwt
from ..RCPs.rcp45 import Emissions as r45e


def Stevens(emissions, stevens_params=np.array([0.001875, 0.634, 60.]),
    ref_isSO2=True, E_pi=0):
    """Calculates aerosol forcing based on Stevens (2015) that relates sulphate
    aerosol forcing to SOx emissions in a logarithmic fashion.

    Input:
        emissions:   anthropogenic emissions database
    Keywords:
        stevens_params: 3 element array
            0. scaling parameter for ERFari (alpha)
            1. scaling parameter for ERFaci (beta)
            2. natural emissions of SOx in Mt/yr
        ref_isSO2:   True if E_SOx_nat is in units of SO2 rather than S.
        E_pi: pre-industrial/reference emissions of SO2 (or S).
    Output:
        ERFari, ERFaci:  aerosol effective radiative forcing due to 
            aerosol-radiation interactions and aerosol-cloud interactions.
    """

    alpha, beta, E_SOx_nat = stevens_params

    factor = 1
    if ref_isSO2:
        factor = molwt.SO2/molwt.S
    em_SOx = emissions[:,5] * factor
    em_pi  = E_pi * factor

    ERFari = -alpha * (em_SOx-em_pi)
#    ERFaci = (
#        (-beta * np.log(em_SOx/E_SOx_nat + 1)) - 
#        (-beta * np.log(em_pi/E_SOx_nat + 1)) )
    ERFaci = (-beta * np.log((em_SOx-em_pi)/E_SOx_nat + 1))
    return ERFari, ERFaci


def aerocom_direct(emissions,
        beta = np.array(
          [-6.2227e-3, 0.0, -3.8392e-4, -1.16551e-3, 1.601537e-2, -1.45339e-3,
           -1.55605e-3]), E_pi=np.zeros(40)
    ):

    """Calculates direct aerosol forcing based on linear relationships between
    emissions and forcing in Aerocom models.

    Reference: Myhre et al., 2013: https://www.atmos-chem-phys.net/13/1853/2013

    If inputs from an RCPs SCEN file are used, the units will be correct.

    Inputs:
        emissions: (nt x 40) emissions array
    Keywords:
        beta: 7-element array of forcing efficiencies in W m-2 (Mt yr-1)-1 for
            SOx, CO, NMVOC, NOx, BC, OC, NH3 (in that order)
        E_pi: pre-industrial emissions (40 element array)
    Outputs:
        Forcing time series
    """

    em_SOx, em_CO, em_NMVOC, em_NOx, em_BC, em_OC, em_NH3 = \
        emissions[:,[5, 6, 7, 8, 9, 10, 11]].T

    F_SOx    = beta[0] * (em_SOx   - E_pi[5])
    F_CO     = beta[1] * (em_CO    - E_pi[6])
    F_NMVOC  = beta[2] * (em_NMVOC - E_pi[7])
    F_NOx    = beta[3] * (em_NOx   - E_pi[8])
    F_BC     = beta[4] * (em_BC    - E_pi[9])
    F_OC     = beta[5] * (em_OC    - E_pi[10])
    F_NH3    = beta[6] * (em_NH3   - E_pi[11])

    ERFari = F_SOx+F_CO+F_NMVOC+F_NOx+F_BC+F_OC+F_NH3

    return ERFari


def ghan_indirect(emissions, fix_pre1850_RCP=True, scale_AR5=False,
    ghan_params=np.array([-1.95011431, 0.01107147, 0.01387492]),
    E_pi=np.zeros(40)):
    """Estimates the aerosol indirect effect based on the simple model in
    Ghan et al., (2013), doi:10.1002/jgrd.50567.

    This function is just an emulator - a full implementation in Python of the
    Ghan routine (originally coded in Fortran) exists, but will require
    optimisation before it can be used in FaIR. I hope to make the full version
    available in a future version.

    A 500-member Latin Hypercube sample of emissions of SOx, NMVOC, BC and OC
    was prepared offline and run through the Ghan simple model and a functional
    relationship fit to the output. SOA aerosol (based on NMVOC emissions) is
    sometimes unreliable and does not exert a strong dependence on the ERF, and
    OC+BC is parameterised as primary organic matter, so the resulting output
    is a function of SOx and (BC+OC) emissions.

    Inputs:
        emissions: (nt x 40) numpy emissions array
    Keywords:
        fix_pre1850_RCP: Use different relationship for 1750/65 to 1850 based
                         on anthropogenic emissions from Skeie et al (2011)
                         for 1750 (atmos-chem-phys.net/11/11827/2011)
        scale_AR5:       If True, scale the forcing output so that the best
                         estimate forcing in 2011 is -0.45 W/m2 based on 2011
                         emissions from the RCPs. The Ghan emulator is built on
                         results from the CAM5 GCM. As reported in AR5 WG1 Ch7,
                         GCMs tend to overestimate forcing from aerosol-cloud
                         interactions.
        ghan_params:     3-element numpy array
                         0: scale factor
                         1: sensitivity to SOx emissions
                         2: sensitivity to BC+OC emissions
    Outputs:
        Forcing timeseries
    """

    year, em_SOx, em_BC, em_OC = emissions[:,[0, 5, 9, 10]].T

    def _ERFaci(em,
        ghan_params=np.array([-1.95011431, 0.01107147, 0.01387492])):
        scale = ghan_params[0]
        b_SOx = ghan_params[1]
        b_POM = ghan_params[2]
        return scale*np.log(1+b_SOx*em[0]+b_POM*em[1])

    # PI forcing was not zero as there were some emissions. Use estimates
    # from Skeie et al, 2011 for 1750 forcing.
    E_1765 = np.array([1.0, 11.2])
    nt = len(year)
    F_pd = np.zeros(nt)
    for i in range(nt):
        if year[i]>=1850 or fix_pre1850_RCP==False:
            F_pd[i] = _ERFaci([em_SOx[i], em_BC[i]+em_OC[i]],
                              ghan_params=ghan_params)
        else:
            # linearly interpolate between 1765 and 1850
            E_1850 = np.array([r45e.sox[85], r45e.bc[85]+r45e.oc[85]])
            F_pd[i] = _ERFaci((year[i]-1765)/85.*E_1850 +
                              (1850-year[i])/85.*E_1765,
                              ghan_params=ghan_params)

    # 1765 emissions = zero forcing
    F_1765 = -0.3002836449793625
    F_2011 = -1.5236182344467388

    # are we rescaling to AR5 best estimate with the default parameters?
    if scale_AR5:
        scale=-0.45/(F_2011-F_1765)
    else:
        scale=1.0

    ERFaci = (F_pd - F_1765) * scale
    return ERFaci


def ghan2(emissions, E_pi, ghan_params):
    """temphack for fair1.6"""
    beta, n_so2, n_pom = ghan_params
    pd_re = -beta * np.log(1 + emissions[:,5]/n_so2 + emissions[:,9:11].sum(axis=1)/n_pom)
    pi_re = -beta * np.log(1 + E_pi[5]/n_so2 + E_pi[9:11].sum()/n_pom)
    return pd_re - pi_re



from __future__ import division

import numpy as np

from ..constants.general import EARTH_RADIUS, SECONDS_PER_YEAR

# This is a stop-gap until I figure out how to couple the openscm-twolayermodel

def forcing_to_temperature(
    temp,
    f0,
    f1,
    lambda_global=1.18,
    ocean_heat_capacity=np.array([8.2, 109.0]),
    ocean_heat_exchange=0.67,
    deep_ocean_efficacy=1.28,
    dt=1):
    """Calculate temperature from a given radiative forcing.

    This follows the forcing to temperature function described in Geoffroy
    et al. (2013a, 2013b).

    Thanks to Glen Harris, UK Met Office, for doing the grunt work.

    Inputs:
        temp: (2,2) numpy array (layer, component) of ocean temperatures
            in timestep t-1 (layer = mixed, deep); (component=fast; slow)
        f0: effective radiative forcing in timestep t-1
        f1: effective radiative forcing in timestep t

    Keywords:
        lambda_global: Climate feedback parameter (convention positive
            stable), W m-2 K-1
        ocean_heat_capacity: 2-element np.ndarray of ocean heat capacities
            (mixed layer, deep ocean) in W yr m-2 K-1
        ocean_heat_exchange: heat exchange coefficient between the two
            ocean layers in W m-2 K-1
        deep_ocean_efficacy: deep ocean efficacy parameter, non-dimensional           
        dt: timestep, year

    Outputs:
        t1: slow and fast contributions to total temperature (2 element array)
        in timestep t
    """

    # care with unit handling. This will be a large number - divide 1e21 for ZJ
    ntoa_joule = 4 * np.pi * EARTH_RADIUS**2 * SECONDS_PER_YEAR
             
    # Define derived constants                
    cdeep_p = ocean_heat_capacity[1] * deep_ocean_efficacy
    gamma_p = ocean_heat_exchange * deep_ocean_efficacy
    g1 = (lambda_global+gamma_p)/ocean_heat_capacity[0]
    g2      = gamma_p/cdeep_p
    g       = g1+g2
    gstar   = g1-g2
    delsqrt = np.sqrt(g*g - 4*g2*lambda_global/ocean_heat_capacity[0])
    afast   = (g + delsqrt)/2
    aslow   = (g - delsqrt)/2
    cc      = 0.5/(ocean_heat_capacity[0]*delsqrt)        
    amix_f  = cc*(gstar+delsqrt)
    amix_s  = -cc*(gstar-delsqrt)        
    adeep_f = -gamma_p/(ocean_heat_capacity[0]*cdeep_p*delsqrt)
    adeep_s = -adeep_f                                            

    temp_mix0 = np.copy(temp[0,:])
    temp_deep0 = np.copy(temp[1,:])

    adf = 1/(afast*dt)
    ads = 1/(aslow*dt)
    exp_f = np.exp(-1./adf)
    exp_s = np.exp(-1./ads)        
    int_f = (f0*adf + f1*(1-adf) - exp_f*(f0*(1+adf)-f1*adf))/afast
    int_s = (f0*ads + f1*(1-ads) - exp_s*(f0*(1+ads)-f1*ads))/aslow

    temp_mix1  = np.array([exp_f*temp_mix0[0] + amix_f*int_f, 
                           exp_s*temp_mix0[1] + amix_s*int_s]) 
                                         
    temp_deep1 = np.array([exp_f*temp_deep0[0] + adeep_f*int_f, 
                           exp_s*temp_deep0[1] + adeep_s*int_s])

    c_dtemp = (
        ocean_heat_capacity[0]*(temp_mix1.sum()-temp_mix0.sum()) + 
        ocean_heat_capacity[1]*(temp_deep1.sum()-temp_deep0.sum())
    )

    heatflux = c_dtemp/dt
    del_ohc  = ntoa_joule * c_dtemp
        
    factor_lambda_eff = (deep_ocean_efficacy-1.0)*ocean_heat_exchange
    if abs(np.sum(temp_mix1)) > 1e-6:
        ratio = (np.sum(temp_mix1) - np.sum(temp_deep1))/np.sum(temp_mix1)
        lambda_eff = lambda_global + factor_lambda_eff*ratio
    else:
        lambda_eff = lambda_global + factor_lambda_eff
        
    return np.array([temp_mix1, temp_deep1]), heatflux, del_ohc, lambda_eff




from __future__ import division

import numpy as np
from ..constants import molwt

def from_aviNOx(emissions, frac, E_ref=2.946, F_ref=0.0448, ref_isNO2=True):
    """Calculates contrail radiative forcing from emissions of aviation NOx.

    Inputs:
        emissions:  Raw emissions data.
        frac:       fraction of total NOx emissions due to aviation.
    Keywords:
        E_ref:      reference-year emissions of aviation NOx, Mt/yr. 2.946 
                    is 2005 emissions of aviation NOx in RCP4.5 measured in
                    Mt NO2/yr.
        F_ref:      Forcing from linear persistent contrails + contrail 
                    induced cirrus. The default of 0.0448 W/m2 for 2005 is
                    from Lee et al, 2009 (Atmos. Environ.,
                    doi:10.1016/j.atmosenv.2009.04.024).
        ref_isNO2:  True if E_ref is in units of NO2 rather than N.
    """

    em_NOx   = emissions[:,8]
    factor = 1

    if ref_isNO2:
        factor = molwt.NO2/molwt.N
    return em_NOx*frac * F_ref/E_ref * factor


def from_fuel(keroseneGt, S_ref=236.063, F_ref=0.0448):
    """Calculates contrail radiative forcing from fuel supply of jet kerosene.

    This method assumes a linear scaling of ERF with kerosene jet fuel
    supplied, which is a proxy for aircraft activity. The relationship may not
    be linear as suggested in IPCC (1999) and depends on many uncertain
    factors.
    IPCC (1999): Aviation and the Global Atmosphere. J.E.Penner, D.H.Lister,
    D.J.Griggs, D.J.Dokken, M.McFarland (Eds.). Cambridge University Press,
    UK. pp 373

    Inputs:
        keroseneGt:  Jet fuel supplied in Gt/yr
    Keywords:
        S_ref:       Jet fuel supplied for a reference year (2005), Gt. Source
                     for default: International Energy Agency, 
                     http://www.iea.org/statistics/statisticssearch/report/
                     ?country=WORLD&product=oil&year=2005
        F_ref:       Forcing from linear persistent contrails + contrail
                     induced cirrus. The default of 0.0448 W/m2 for 2005 is
                     from Lee et al, 2009 (Atmos. Environ.,
                     doi:10.1016/j.atmosenv.2009.04.024).
    """

    return keroseneGt * F_ref/S_ref



from __future__ import division

import numpy as np
from itertools import compress
from ..constants.general import M_ATMOS
from ..constants import molwt as molwt_builtin, preindconc,\
   lifetime as lt_builtin


def _lookup(species):
    # get list of GHGs from preindconc model
    named_gases = dir(preindconc)

    # strip out reserved names and the "aslist" list
    include = [i[:2]!='__' and i!='aslist' for i in named_gases]
    named_gases = list(compress(named_gases, include))

    # see if the species appears on the list of builtin gases and return
    # properties if so, otherwise raise error
    if species in named_gases:
        inout = {}
        inout['pi'] = preindconc
        inout['lt'] = lt_builtin
        inout['mw'] = molwt_builtin
        exec("C, L, M = pi."+species+", lt."+species+", mw."+species, inout)
        return inout['C'], inout['L'], inout['M']
    else:
        raise ValueError(species + ' is not in the list of recognised '+
          'greenhouse gases') 


def emissions(C=None, lifetime=None, molwt=None, species=None):
    """Calculate steady state background emissions from a given lifetime

    Keywords:
        C: concentrations, scalar or None
            if scalar, steady state concentrations to acheive.
            if None, use the pre-industrial concentrations for the specified
              gas.
        lifetime: scalar or None
            if scalar, use the given greenhouse gas atmospheric lifetime
            if None, use the default lifetime for the specified gas
        molwt: scalar or None
            if scalar, use molecular weight of given species
            if None, use the default molecular weight for the specified gas
        species: string or None
            Name of the greenhouse gas concentrations you want to calculate.
            See ..constants.lifetime module for used names.
            If None, use the values given in C, lifetime and wt.

    Any user-specified values for C, lifetime and molwt overrides the
    defaults. They must be specified if species is None.

    Returns:
        emissions (scalar)
    """

    # if not using a built-in gas, C, lifetime and molwt must be specified.
    if species is None:
        if any((C is None, lifetime is None, molwt is None)):
            raise ValueError('If species is not given then C, '+
              'lifetime and molwt must be specified.')
    else:
        # populate defaults but override if values given
        species = species.upper()
        C0, lifetime0, molwt0 = _lookup(species)
        if C is None: C=C0
        if lifetime is None: lifetime=lifetime0
        if molwt is None: molwt=molwt0

        # convert units for N2O
        if species == 'N2O':
            molwt = molwt * molwt_builtin.N2/molwt_builtin.N2O

    # now invert the emissions-concentrations relationship
    E = (C * (1.0 - np.exp(-1.0 / lifetime)) * M_ATMOS * molwt /
      molwt_builtin.AIR) * 1e-18

    return E



from __future__ import division

import datetime as dt

import os
import numpy as np
import pandas as pd
from scipy.interpolate import interp1d

from ..constants import molwt

try:
    from scmdata import ScmDataFrame, ScmRun
    has_scmdata = True
except ImportError:
    has_scmdata = False


class SSP245WorldEmms:
    def __init__(self):
        self._loaded = False
        self._loaded_fair_history = False

    @property
    def values(self):
        if not self._loaded:
            self._values = (
                ScmRun(
                    os.path.join(
                        os.path.dirname(__file__),
                        '../SSPs/data/rcmip-emissions-annual-means-4-0-0-ssp-only.csv'),
                        lowercase_cols=True
                )
                .filter(scenario="ssp245", region="World", variable="Emissions*")
            )
            self._values = self._values.interpolate([dt.datetime(y, 1, 1) for y in self._values["year"]])

        self._loaded = True

        return self._values

    @property
    def values_fair_units(self):
        if not self._loaded_fair_history:
            ssp_df_hist = self.values
            for variable in ssp_df_hist.get_unique_meta("variable"):
                in_unit = ssp_df_hist.filter(variable=variable).get_unique_meta("unit", no_duplicates=True)

                try:
                    _, fair_unit, context = _get_fair_col_unit_context(variable)
                except AssertionError:
                    # FaIR does not model the variable
                    assert variable in [
                        'Emissions|F-Gases|HFC|HFC152a',
                        'Emissions|F-Gases|HFC|HFC236fa',
                        'Emissions|F-Gases|HFC|HFC365mfc',
                        'Emissions|F-Gases|NF3',
                        'Emissions|F-Gases|PFC|C3F8',
                        'Emissions|F-Gases|PFC|C4F10',
                        'Emissions|F-Gases|PFC|C5F12',
                        'Emissions|F-Gases|PFC|C7F16',
                        'Emissions|F-Gases|PFC|C8F18',
                        'Emissions|F-Gases|PFC|cC4F8',
                        'Emissions|F-Gases|SO2F2',
                        'Emissions|Montreal Gases|CH2Cl2',
                        'Emissions|Montreal Gases|CHCl3',
                    ]

                    continue

                if in_unit != fair_unit:
                    ssp_df_hist = ssp_df_hist.convert_unit(fair_unit, variable=variable, context=context)

            self._values_fair_history = ssp_df_hist

        self._loaded_fair_history = True

        return self._values_fair_history



# TODO: lazy load this and only load once
ssp245_world_emms_holder = SSP245WorldEmms()


EMISSIONS_SPECIES_UNITS_CONTEXT = pd.DataFrame((  # in fair 1.6, order is important
        ('|CO2|MAGICC Fossil and Industrial', 'GtC / yr', None),
        ('|CO2|MAGICC AFOLU', 'GtC / yr', None),
        ('|CH4', 'MtCH4 / yr', None),
        ('|N2O', 'MtN / yr', None),
        ('|Sulfur', 'MtS / yr', None),
        ('|CO', 'MtCO / yr', None),
        ('|VOC', 'MtNMVOC / yr', None),
        ('|NOx', 'MtN / yr', "NOx_conversions"),
        ('|BC', 'MtBC / yr', None),
        ('|OC', 'MtOC / yr', None),
        ('|NH3', 'MtN / yr', None),
        ('|CF4', 'ktCF4 / yr', None),
        ('|C2F6', 'ktC2F6 / yr', None),
        ('|C6F14', 'ktC6F14 / yr', None),
        ('|HFC23', 'ktHFC23 / yr', None),
        ('|HFC32', 'ktHFC32 / yr', None),
        ('|HFC4310mee', 'ktHFC4310mee / yr', None),
        ('|HFC125', 'ktHFC125 / yr', None),
        ('|HFC134a', 'ktHFC134a / yr', None),
        ('|HFC143a', 'ktHFC143a / yr', None),
        ('|HFC227ea', 'ktHFC227ea / yr', None),
        ('|HFC245fa', 'ktHFC245fa / yr', None),
        ('|SF6', 'ktSF6 / yr', None),
        ('|CFC11', 'ktCFC11 / yr', None),
        ('|CFC12', 'ktCFC12 / yr', None),
        ('|CFC113', 'ktCFC113 / yr', None),
        ('|CFC114', 'ktCFC114 / yr', None),
        ('|CFC115', 'ktCFC115 / yr', None),
        ('|CCl4', 'ktCCl4 / yr', None),
        ('|CH3CCl3', 'ktCH3CCl3 / yr', None),
        ('|HCFC22', 'ktHCFC22 / yr', None),
        ('|HCFC141b', 'ktHCFC141b / yr', None),
        ('|HCFC142b', 'ktHCFC142b / yr', None),
        ('|Halon1211', 'ktHalon1211 / yr', None),
        ('|Halon1202', 'ktHalon1202 / yr', None),
        ('|Halon1301', 'ktHalon1301 / yr', None),
        ('|Halon2402', 'ktHalon2402 / yr', None),
        ('|CH3Br', 'ktCH3Br / yr', None),
        ('|CH3Cl', 'ktCH3Cl / yr', None),
    ),
    columns=["species", "in_unit", "context"],
)


def _get_fair_col_unit_context(variable):
    row = EMISSIONS_SPECIES_UNITS_CONTEXT["species"].apply(lambda x: variable.endswith(x))

    in_unit = EMISSIONS_SPECIES_UNITS_CONTEXT[row]["in_unit"]

    assert in_unit.shape[0] == 1, in_unit

    fair_col = int(row[row].index.values) + 1  # first col is time
    in_unit = in_unit.iloc[0]
    context = EMISSIONS_SPECIES_UNITS_CONTEXT[row]["context"].iloc[0]

    return fair_col, in_unit, context


def scmdf_to_emissions(scmdf, include_cfcs=True, startyear=1765, endyear=2100):
    """
    Opens an ScmDataFrame and extracts the data. Interpolates linearly
    between non-consecutive years in the SCEN file. Fills in chlorinated gases
    from a specified SSP scenario.

    Note this is a temporary fix for FaIR 1.6.

    Inputs:
        scmdf: ScmDataFrame

    Keywords:
        include_cfcs: bool
            MAGICC files do not come loaded with CFCs (indices 24-39).
            - if True, use the values from RCMIP for SSPs (all scenarios are
                the same).
            - Use False to ignore and create a 23-species emission file.
        startyear: First year of output file.
        endyear: Last year of output file.

    Returns:
        nt x 40 numpy emissions array (nt x 23 if ``include_cfcs`` is ``False``)
    """

    # We expect that aeneris and silicone are going to give us a nicely
    # formatted ScmDataFrame with all 23 species present and correct at
    # timesteps 2015, 2020 and ten-yearly to 2100.
    # We also implicitly assume that data up until 2014 will follow SSP
    # historical.
    # This adapter will not be tested on anything else!

    n_cols = 40
    nt = endyear - startyear + 1

    data_out = np.ones((nt, n_cols)) * np.nan
    data_out[:,0] = np.arange(startyear, endyear+1)

    if not has_scmdata:
        raise ImportError("This is not going to work without having scmdata installed")

    if not isinstance(scmdf, ScmDataFrame):
        raise TypeError("scmdf must be an scmdata.ScmDataFrame instance")

    if not include_cfcs:
        raise NotImplementedError("include_cfcs equal to False")

    if scmdf[["model", "scenario"]].drop_duplicates().shape[0] != 1:
        raise AssertionError("Should only have one model-scenario pair")

    scen_start_year = 2015

    scmdf = ScmRun(scmdf.timeseries()).interpolate(
        [dt.datetime(y, 1, 1) for y in range(scen_start_year, endyear + 1)]
    )

    years = scmdf["year"].values
    first_scenyear = years[0]
    first_scen_row = int(first_scenyear-startyear)

    # if correct units and interpolation were guaranteed we could do this for scenario too which is quicker
    hist_df = ssp245_world_emms_holder.values_fair_units.filter(
        year=range(startyear, 2015)
    ).timeseries()

    future_ssp245_df = ssp245_world_emms_holder.values_fair_units.filter(
        year=range(2015, endyear + 1)
    ).timeseries()

    for species in EMISSIONS_SPECIES_UNITS_CONTEXT["species"]:
        fair_col, _, _ = _get_fair_col_unit_context(species)

        hist_df_row = hist_df.index.get_level_values("variable").str.endswith(species)

        data_out[: first_scen_row, fair_col] = hist_df[hist_df_row].values.squeeze()

        future_ssp245_df_row = future_ssp245_df.index.get_level_values("variable").str.endswith(species)

        data_out[first_scen_row :, fair_col] = future_ssp245_df[future_ssp245_df_row].values.squeeze()


    for var_df in scmdf.groupby("variable"):
        variable = var_df.get_unique_meta("variable", no_duplicates=True)
        in_unit = var_df.get_unique_meta("unit", no_duplicates=True)
        fair_col, fair_unit, context = _get_fair_col_unit_context(variable)

        if in_unit != fair_unit:
            var_df_fair_unit = var_df.convert_unit(fair_unit, context=context)
        else:
            var_df_fair_unit = var_df

        data_out[first_scen_row :, fair_col] = var_df_fair_unit.values.squeeze()


    return data_out



# Convenience module for loading in RCP emissions datasets
#
# Usage:
#
# import rcp85
# rcp85.Emissions.co2

import numpy as np
import os
emissions_filename = os.path.join(
    os.path.dirname(__file__), 'data/RCP85_EMISSIONS.csv')
concentrations_filename = os.path.join(
    os.path.dirname(__file__), 'data/RCP85_MIDYEAR_CONCENTRATIONS.csv')
forcing_filename = os.path.join(
    os.path.dirname(__file__), 'data/RCP85_MIDYEAR_RADFORCING.csv')
aviNOx_filename = os.path.join(
    os.path.dirname(__file__), 'data/aviNOx_fraction.csv')
fossilCH4_filename = os.path.join(
    os.path.dirname(__file__), 'data/fossilCH4_fraction.csv')

aviNOx_frac = np.loadtxt(aviNOx_filename, skiprows=5, usecols=(4,),
    delimiter=',')
fossilCH4_frac = np.loadtxt(fossilCH4_filename, skiprows=5, usecols=(4,),
    delimiter=',')

class Emissions:
    emissions = np.loadtxt(emissions_filename, skiprows=37, delimiter=',')
    year      = emissions[:,0]
    co2_fossil= emissions[:,1]
    co2_land  = emissions[:,2]
    co2       = np.sum(emissions[:,1:3],axis=1)
    ch4       = emissions[:,3]
    n2o       = emissions[:,4]
    sox       = emissions[:,5]
    co        = emissions[:,6]
    nmvoc     = emissions[:,7]
    nox       = emissions[:,8]
    bc        = emissions[:,9]
    oc        = emissions[:,10]
    nh3       = emissions[:,11]
    cf4       = emissions[:,12]
    c2f6      = emissions[:,13]
    c6f14     = emissions[:,14]
    hfc23     = emissions[:,15]
    hfc32     = emissions[:,16]
    hfc43_10  = emissions[:,17]
    hfc125    = emissions[:,18]
    hfc134a   = emissions[:,19]
    hfc143a   = emissions[:,20]
    hfc227ea  = emissions[:,21]
    hfc245fa  = emissions[:,22]
    sf6       = emissions[:,23]
    cfc11     = emissions[:,24]
    cfc12     = emissions[:,25]
    cfc113    = emissions[:,26]
    cfc114    = emissions[:,27]
    cfc115    = emissions[:,28]
    carb_tet  = emissions[:,29]
    mcf       = emissions[:,30]
    hcfc22    = emissions[:,31]
    hcfc141b  = emissions[:,32]
    hcfc142b  = emissions[:,33]
    halon1211 = emissions[:,34]
    halon1202 = emissions[:,35]
    halon1301 = emissions[:,36]
    halon2402 = emissions[:,37]
    ch3br     = emissions[:,38]
    ch3cl     = emissions[:,39]


class Concentrations:
    concentrations = np.loadtxt(concentrations_filename, skiprows=39,
      delimiter=',')
    gas_indices= np.concatenate(([3,4,5], np.arange(8,36)))
    gases      = concentrations[:,gas_indices]
    year       = concentrations[:,0]
    co2eq      = concentrations[:,1]
    kyotoco2eq = concentrations[:,2]
    co2        = concentrations[:,3]
    ch4        = concentrations[:,4]
    n2o        = concentrations[:,5]
    fgassum    = concentrations[:,6]
    mhalosum   = concentrations[:,7]
    cf4        = concentrations[:,8]
    c2f6       = concentrations[:,9]
    c6f14      = concentrations[:,10]
    hfc23      = concentrations[:,11]
    hfc32      = concentrations[:,12]
    hfc43_10   = concentrations[:,13]
    hfc125     = concentrations[:,14]
    hfc134a    = concentrations[:,15]
    hfc143a    = concentrations[:,16]
    hfc227ea   = concentrations[:,17]
    hfc245fa   = concentrations[:,18]
    sf6        = concentrations[:,19]
    cfc11      = concentrations[:,20]
    cfc12      = concentrations[:,21]
    cfc113     = concentrations[:,22]
    cfc114     = concentrations[:,23]
    cfc115     = concentrations[:,24]
    carb_tet   = concentrations[:,25]
    mcf        = concentrations[:,26]
    hcfc22     = concentrations[:,27]
    hcfc141b   = concentrations[:,28]
    hcfc142b   = concentrations[:,29]
    halon1211  = concentrations[:,30]
    halon1202  = concentrations[:,31]
    halon1301  = concentrations[:,32]
    halon2402  = concentrations[:,33]
    ch3br      = concentrations[:,34]
    ch3cl      = concentrations[:,35]    

    
class Forcing:
    forcing   = np.loadtxt(forcing_filename, skiprows=59, delimiter=',')
    year      = forcing[:,0]
    total     = forcing[:,1]
    volcanic  = forcing[:,2]
    solar     = forcing[:,3]
    ghg       = forcing[:,5]
    co2       = forcing[:,8]
    ch4       = forcing[:,9]
    n2o       = forcing[:,10]
    fgas      = forcing[:,11]
    halo      = forcing[:,12]
    aero      = forcing[:,41]
    cloud     = forcing[:,48]
    strato3   = forcing[:,49]
    tropo3    = forcing[:,50]
    stwv      = forcing[:,51]
    dust      = forcing[:,47]
    landuse   = forcing[:,52]
    bcsnow    = forcing[:,53]




# pre-industrial GHG concentrations in their default units.
# Source: http://www.pik-potsdam.de/~mmalte/rcps. Year 1765 in all cases. 
# TODO: needs to be updated for AR6/CMIP6 based on Meinshausen et al., 2017
CO2         = 278.05158  # ppm
CH4         = 721.89411  # ppb
N2O         = 272.95961  # ppb
CF4         = 35.        # ppt
C2F6        = 0.         # ppt
C3F8        = 0.         # ppt
C4F10       = 0.         # ppt
C5F12       = 0.         # ppt
C6F14       = 0.         # ppt
C7F16       = 0.         # ppt
C8F18       = 0.         # ppt
C_C4F8      = 0.         # ppt
HFC23       = 0.         # ppt
HFC32       = 0.         # ppt
HFC43_10    = 0.         # ppt
HFC43_10MEE = 0.         # ppt
HFC125      = 0.         # ppt
HFC134A     = 0.         # ppt
HFC143A     = 0.         # ppt
HFC152A     = 0.         # ppt
HFC227EA    = 0.         # ppt
HFC236FA    = 0.         # ppt
HFC245FA    = 0.         # ppt
HFC365MFC   = 0.         # ppt
SF6         = 0.         # ppt
NF3         = 0.         # ppt
SO2F2       = 0.         # ppt
CFC11       = 0.         # ppt
CFC12       = 0.         # ppt
CFC113      = 0.         # ppt
CFC114      = 0.         # ppt
CFC115      = 0.         # ppt
CARB_TET    = 0.         # ppt
CCL4        = CARB_TET   # ppt
MCF         = 0.         # ppt
CH3CCL3     = MCF        # ppt
HCFC22      = 0.         # ppt
HCFC141B    = 0.         # ppt
HCFC142B    = 0.         # ppt
HALON1211   = 0.         # ppt
HALON1202   = 0.         # ppt
HALON1301   = 0.         # ppt
HALON2402   = 0.         # ppt
CH3BR       = 5.8        # ppt
CH3CL       = 480.       # ppt
CH2CL2      = 0.         # ppt
CHCL3       = 0.         # ppt

# This is the list of gases included in the RCPs/AR5/CMIP5.
aslist   = [CO2, CH4, N2O, CF4, C2F6, C6F14, HFC23, HFC32, HFC43_10, HFC125,
            HFC134A, HFC143A, HFC227EA, HFC245FA, SF6, CFC11, CFC12, CFC113,
            CFC114, CFC115, CARB_TET, MCF, HCFC22, HCFC141B, HCFC142B,
            HALON1211, HALON1202, HALON1301, HALON2402, CH3BR, CH3CL]



from . import molwt

M_ATMOS          = 5.1352e18 # mass of atmosphere, kg
EARTH_RADIUS     = 6371000   # m
SECONDS_PER_YEAR = 60 * 60 * 24 * 365.24219 # Length of tropical year
    # https://en.wikipedia.org/wiki/Tropical_year

# Conversion between ppm CO2 and GtC emissions
ppm_gtc   = M_ATMOS/1e18*molwt.C/molwt.AIR



import numpy as np
import os

forcing_filename = os.path.join(os.path.dirname(__file__), 'cmip6_solar.csv')

class Forcing:
    forcing  = np.loadtxt(forcing_filename, skiprows=7, delimiter=',')
    year     = forcing[:,0]
    solar    = forcing[:,1]



# usage
# from fair.ancil import historical_scaling
# s = historical_scaling.all

import numpy as np
import os

filename = os.path.join(os.path.dirname(__file__), 'historical_scaling.csv')
all = np.loadtxt(filename, skiprows=0, delimiter=',')
co2 = all[:,0]



from __future__ import division

import numpy as np
import scipy.stats as st
import os
from functools import reduce

def mvlognorm(data, n=1000, seed=None, correlated=True):
    """Returns joint lognormal random variables.
    
    Inputs:
        data: (m, p) array of 'observations' where m is the number
            of observations and p is the number of parameters to
            estimate.
        
    Keywords:
        n: number of 
        seed: random seed for variable generation
        correlated: logical. If True, assume random variables are
            correlated, and calculate the correlation coefficient.
        
    Outputs:
        (n, p) array of simulated joint lognormal random variables.
        
    Method: This function takes the input data (untransformed) and 
    calculates the mean, standard deviation and correlation coefficient
    of the data, transforms it to a lognormal, and returns simulated
    data based on the input 'observations'.
    
    It is based on the MethylCapSig R package by Deepak N. Ayyala et al.,
    https://CRAN.R-project.org/package=MethylCapSig
    """
    
    p = data.shape[1]
    mu = np.mean(data, axis=0)
    Sigma = np.var(data, axis=0)
    if correlated:
        corr = np.corrcoef(np.log(data), rowvar=False)
    else:
        corr = np.eye((p))
    
    alpha = np.log(mu) - 0.5*np.log(1.0 + Sigma/mu**2)
    beta = np.diag(np.log(1.0 + Sigma/(mu**2)))
    
    delta = reduce(np.matmul, [np.sqrt(beta), corr, np.sqrt(beta)])
    delta_eigenvalues, delta_eigenvectors = np.linalg.eig(delta)
    root_delta = reduce(np.matmul, 
      [delta_eigenvectors, np.diag(np.sqrt(delta_eigenvalues)), delta_eigenvectors.T])

    out = st.norm.rvs(size=(n,p), random_state=seed)
    for i in range(n):
        out[i,:] = np.exp(alpha[:] + np.matmul(root_delta, out[i,:]))

    return out


def tcrecs_generate(tcrecs_in='cmip5', dist='lognorm', n=1000, correlated=True,
                    strip_ecs_lt_tcr=True,
                    seed=None):
    """Generates a distribution of TCR and ECS.
    
    Inputs:
        tcrecs_in: either 'cmip5' for pre-shipped CMIP5 TCR (transient climate
        response) and ECS (equilibrium climate sensitivity) values, or a 
        2-column array of TCR and ECS values to sample from.
        
    Keywords:
        dist: string. Distribution to use when constructing the
            joint distribution. Accepted values are norm and 
            lognorm (default).
        n: number of samples to generate. Default 1000.
        correlated: logical. If True (default), assume ECS and TCR
            inputs are correlated. The function calculates the
            correlation coefficient automatically.
        strip_ecs_lt_tcr: logical. If True (default), remove values
            where ECS < TCR in place (but still return n samples.)
        seed: random seed for generating variables.

    Output:
        (n, 2) array of sampled ECS, TCR pairs."""
    
    if type(tcrecs_in) is str and tcrecs_in=='cmip5':
        filepath = os.path.join(os.path.dirname(__file__),
          'tcrecs/cmip5tcrecs.csv')
        tcrecs_in = np.loadtxt(filepath, delimiter=',', skiprows=3)
    try:
        assert(type(tcrecs_in) is np.ndarray)
        assert(tcrecs_in.ndim == 2)
        assert(tcrecs_in.shape[1] == 2)
    except AssertionError:
        raise ValueError('tcrecs_in should "cmip5" or an array of shape (n, 2)')
    
    dist = dist.lower()
    
    def _genvar(tcrecs, dist, n, seed, correlated):
        if dist=='lognorm':
            out = mvlognorm(tcrecs, n=n, seed=seed, correlated=correlated)
        elif dist=='norm':
            mu = np.mean(tcrecs, axis=0)
            if correlated:
                cov = np.cov(tcrecs, rowvar=False)
            else:
                cov = np.diag(np.var(tcrecs, axis=0))
            out = st.multivariate_normal.rvs(mu, cov, size=n, random_state=seed)
        else:
            raise ValueError('dist should be "norm" or "lognorm"')
        return out

    tcrecs_out = _genvar(tcrecs_in, dist, n, seed, correlated)
    
    if strip_ecs_lt_tcr:
        tcrecs_out = np.delete(tcrecs_out, 
          np.where(tcrecs_out[:,0] > tcrecs_out[:,1]), axis=0)
        nreq = n - len(tcrecs_out[:,0])
        while nreq>0:
            # required otherwise we are repeating values
            # this is still deterministic for constant ensemble size
            if seed is not None:
                seed = seed + nreq
            new = _genvar(tcrecs_out, dist, n-len(tcrecs_out[:,0]),
              seed, correlated)
            tcrecs_out = np.append(tcrecs_out, new, axis=0)
            tcrecs_out = np.delete(tcrecs_out, np.where(tcrecs_out[:,0] > 
              tcrecs_out[:,1]), axis=0)
            nreq = n - len(tcrecs_out[:,0])
    return tcrecs_out



from __future__ import division

import numpy as np

from ..constants.general import ppm_gtc

"""Gas cycle functions from Generalised Impulse Response Model v1.0.0.

Much of this has been adapted from:

Leach et al., 2020, Geoscientific Model Development
https://www.geosci-model-dev-discuss.net/gmd-2019-379/
"""

def calculate_alpha(cumulative_emissions,airborne_emissions,temperature,r0,rC,rT,g0,g1,iirf_max = 97.0):
    """
    Calculate CO2 time constant scaling factor.

    Inputs:
        cumulative_emissions: GtC cumulative emissions since pre-industrial.
        airborne_emissions: GtC total emissions remaining in the atmosphere.
        temperature: K temperature anomaly since pre-industrial.
        r0: pre-industrial 100-year time-integrated airborne fraction.
        rC: sensitivity of 100-year time-integrated airborne fraction with 
            atmospheric carbon stock.
        rT: sensitivity of 100-year time-integrated airborne fraction with
            temperature anomaly.
        g0: parameter for alpha
        g1: parameter for alpha
    Keywords:
        iirf_max: maximum allowable value to 100-year time-integrated airborne
            fraction
    Outputs:
        alpha: scaling factor.
    """
    iirf = r0 + rC * (cumulative_emissions-airborne_emissions) + rT * temperature
    iirf = (iirf>iirf_max) * iirf_max + iirf * (iirf<iirf_max)
    alpha = g0 * np.sinh(iirf / g1)
    return alpha

def step_concentration(carbon_boxes0,emissions,alpha,a,tau,Cpi,dt=1):
    """
    Calculate concentrations from emissions.

    Inputs:
        carbon_boxes0: CO2 boxes at the end of the previous timestep.
        emissions: GtC CO2 emissions this timestep.
        alpha: CO2 time constant scaling factor.
        a: CO2 partitioning coefficient
        tau: CO2 atmospheric time constants (unscaled).
        Cpi: pre-industrial CO2 concentrations (ppm).
    Keywords:
        dt: timestep in years.
    Outputs:
        C: CO2 concentration (ppm)
        carbon_boxes1: CO2 boxes at the end of this timestep.
        airbone_emissions: GtC total emissions remaining in atmosphere.
    """
    carbon_boxes1 = emissions / ppm_gtc * a * alpha * (tau/dt) * (
        1. - np.exp(-dt/(alpha*tau))) + carbon_boxes0 * np.exp(-dt/(alpha*tau))
    C = Cpi + np.sum(carbon_boxes1 + carbon_boxes0) / 2
    airborne_emissions = np.sum(carbon_boxes1) * ppm_gtc
    return C, carbon_boxes1, airborne_emissions



import numpy as np

"""Default values of carbon cycle parameters.

r0      : pre-industrial time-integrated airborne fraction of CO2 (yr)
rc      : change in time-integrated airborne fraction with CO2 (yr/GtC)
rt      : change in time-integrated airborne fraction with temperature (yr/K)
iirf_h  : time horizon for time integrated airborne fraction (yr)
iirf_max: maximum allowed value of iirf (yr)
a       : partition coefficient of carbon boxes
tau     : present-day decay time constants of CO2 (yr)
"""

r0       = 35.0
rc       = 0.019
rt       = 4.165
iirf_h   = 100.0
iirf_max = 97.0
a        = np.array([0.2173,0.2240,0.2824,0.2763])
tau      = np.array([1000000,394.4,36.54,4.304])


from __future__ import division

import numpy as np
from ..constants import molwt, radeff
from ..constants.general import M_ATMOS

def gwp(H, tau, re, wt, f=0., a=np.array([0.2173, 0.2240, 0.2824, 0.2763]),
    tau_co2=np.array([np.inf, 394.4, 36.54, 4.304])):

    """Calculates Global Warming Potentials

    Inputs:
        H: Time horizon (years)
        tau: lifetime of gas
        re: radiative efficiency of gas
        wt: molecular weight of gas

    Keywords:
        f: feedback factor. f=0.65 for methane and -0.071874 for N2O
        a: impulse response partition for CO2
        tau_co2: lifetimes for impulse response partition for CO2

    Outputs:
        global warming potential, scalar

    TODO: recognise gas from list and select tau, re, wt, f accordingly
    """

    agwp_co2 = radeff.CO2 * (
      a[0]*H + np.sum(a[1:]*tau_co2[1:]*(1-np.exp(-H/tau_co2[1:])))) * (
      molwt.AIR/molwt.CO2 * 1e9/M_ATMOS)

    agwp_gas = re * (1.0+f) * tau * (1.0 - np.exp(-H/tau)) * (
      molwt.AIR/wt * 1e9/M_ATMOS)

    gwp = agwp_gas/agwp_co2
    return gwp



from __future__ import division

import numpy as np
from scipy.optimize import root
from .gas_cycle.fair1 import _iirf_simple, _iirf_interp
from .forcing.ghg import co2_log
from .defaults import carbon, thermal
from .constants import molwt
from .constants.general import ppm_gtc
from .temperature.millar import forcing_to_temperature, calculate_q


def infer_emissions(e1, c1_prescribed, carbon_boxes0, tau_new, a, c_pi):
    """Matches prescribed concentrations to forward-calculated concentrations.

    Inputs:
        e1            : emissions in timestep t, GtC
        c1_prescribed : CO2 concentrations in timestep t, ppmv
        carbon_boxes0 : carbon stored in each atmospheric reservoir at timestep
                        t-1 (GtC)
        tau_new       : decay time constants of CO2 (yr)
        a             : partition coefficient of carbon boxes
        c_pi          : pre-industrial concentration of CO2, ppmv
    """
    
    c1_calculated = np.sum(carbon_boxes0*np.exp(-1.0/tau_new) + a*e1 / ppm_gtc) + c_pi
    return c1_calculated-c1_prescribed


def inverse_carbon_cycle(c1, c_acc0, temp, r0, rc, rt, iirf_max, time_scale_sf,
                         a, tau, iirf_h, carbon_boxes0, c_pi, c0, e0):
    """Calculates CO2 emissions from concentrations.
    
    Inputs:
        c1            : concentration of CO2 in timestep t, ppmv
        c_acc0        : cumulative airborne carbon anomaly (GtC) since
                        pre-industrial, timestep t-1
        temp          : temperature anomaly above pre-industrial (K)
        r0            : pre-industrial time-integrated airborne fraction (yr)
        rc            : sensitivity of time-integrated airborne fraction to 
                        airborne carbon (yr/GtC)
        rt            : sensitivity of time-integrated airborne fraction to
                        temperature (yr/K)
        iirf_max      : maximum value of time-integrated airborne fraction (yr)
        time_scale_sf0: initial guess of alpha scaling factor
        a             : partition coefficient of carbon boxes
        tau           : present-day decay time constants of CO2 (yr)
        iirf_h        : time horizon for time-integrated airborne fraction (yr)
        carbon_boxes0 : carbon stored in each atmospheric reservoir at timestep
                        t-1 (GtC)
        c_pi          : pre-industrial concentration of CO2, ppmv
        c0            : concentration of CO2 in timestep t-1, ppmv
        e0            : emissions of CO2 in timestep t, GtC

    Outputs:
        e1            : emissions of CO2 in timestep t, GtC
        c_acc1        : cumulative airborne carbon anomaly (GtC) since
                        pre-industrial, timestep t
        carbon_boxes1 : carbon stored in each atmospheric reservoir at timestep
                        t (GtC)
        time_scale_sf : scale factor for CO2 decay constants
    """
     

    iirf = _iirf_simple(c_acc0, temp, r0, rc, rt, iirf_max)
    time_scale_sf = root(_iirf_interp, time_scale_sf,
      args=(a, tau, iirf_h, iirf))['x']
    tau_new = tau * time_scale_sf
    e1 = root(infer_emissions, e0, args=(c1, carbon_boxes0, tau_new, a, c_pi))['x']
    c_acc1 = c_acc0 + 0.5*(e1 + e0) - (c1 - c0)*ppm_gtc
    carbon_boxes1 = carbon_boxes0*np.exp(-1.0/tau_new) + a*e1 / ppm_gtc
    return e1, c_acc1, carbon_boxes1, time_scale_sf

    
def inverse_fair_scm(
# Longer term goal: one calling interface only - appropriate function
# determined from call
    C             = None,
    other_rf      = 0.0,
    q             = thermal.q,
    tcrecs        = thermal.tcrecs,
    d             = thermal.d,
    F2x           = thermal.f2x,
    tcr_dbl       = thermal.tcr_dbl,
    a             = carbon.a,
    tau           = carbon.tau,
    r0            = carbon.r0,
    rc            = carbon.rc,
    rt            = carbon.rt,
    iirf_max      = carbon.iirf_max,
    iirf_h        = carbon.iirf_h,
    C_pi          = 278.,
    time_scale_sf = 0.16,
    restart_in    = False,
    restart_out   = False,
    ):

    """Diagnoses emissions from prescribed concentrations.

    Inputs:
        C             : concentrations of CO2, ppmv
        other_rf      : non-CO2 radiative forcing (scalar or numpy array, W/m2)
        q             : coefficients of slow and fast temperature change.
                        Overridden if tcrecs is specified.
        tcrecs        : transient climate response and equilibrium climate
                        sensitivity array (2-element or (nt, 2))
        d             : timescales of slow and fast contribution to temperature
                        change
        F2x           : radiative forcing from a doubling of CO2 concentrations
                        (W/m2)
        tcr_dbl       : timescale over which a 1% compound increase of CO2 acts
                        (yr)
        a             : partition fractions for CO2 boxes
        tau           : time constants for CO2 boxes
        r0            : pre-industrial time-integrated airborne fraction (yr)
        rc            : sensitivity of time-integrated airborne fraction to
                        airborne carbon (yr/GtC)
        rt            : sensitivity of time-integrated airborne fraction to
                        temperature (yr/K)
        iirf_max      : maximum value of time-integrated airborne fraction (yr)
        iirf_h        : time horizon for time-integrated airborne fraction
        C_pi          : pre-industrial concentration of CO2, ppmv
        time_scale_sf : initial guess for scaling factor of CO2 time constants.
                        Overridden if using a restart.
        restart_in    : Allows a restart of the carbon cycle from a non-initial
                        state. A 6-tuple of:
                          array of accumulated carbon in each atmospheric box,
                          array of slow and fast temperature contributions,
                          total accumulated carbon,
                          emissions in the timestep before restart
                          time constant scale factor
                          CO2 concentrations in the timestep before restart
        restart_out   : if True, return the restart state as an extra output.
                        See restart_in.
    Outputs:
        E             : Timeseries of diagnosed CO2 emissions in GtC
        F             : Timeseries of total radiative forcing, W/m2
        T             : Timeseries of temperature anomaly since pre-industrial
        restart       : if restart_out=True, 6-tuple of carbon cycle state
                        parameters. See restart_in.
    """
    
    # Error checking and validation goes here...

    # Dimensions
    nt = len(C)
    carbon_boxes_shape = (nt, a.shape[0])
    thermal_boxes_shape = (nt, d.shape[0])
    
    # Thermal response
    if type(tcrecs) is np.ndarray:
        q     = calculate_q(tcrecs, d, F2x, tcr_dbl, nt)
    
    # Allocate intermediate and output arrays
    C_acc     = np.zeros(nt)
    R_i       = np.zeros(carbon_boxes_shape)
    emissions = np.zeros(nt)
    T_j       = np.zeros(thermal_boxes_shape)
    F         = np.zeros(nt)
    
    if np.isscalar(other_rf):
        other_rf = other_rf * np.ones(nt)
    
    # First timestep
    if restart_in:
        R_i_minus1    = restart_in[0]
        T_j_minus1    = restart_in[1]
        C_acc_minus1  = restart_in[2]
        E_minus1      = restart_in[3]
        time_scale_sf = restart_in[4]
        C_minus1      = restart_in[5]
        emissions[0], C_acc[0], R_i[0,:], time_scale_sf = (
            inverse_carbon_cycle(
                C[0], C_acc_minus1, np.sum(T_j_minus1), r0, rc, rt, iirf_max,
                time_scale_sf, a, tau, iirf_h, R_i_minus1,
                C_pi, C_minus1, E_minus1
            )
        )
        F[0]          = co2_log(C[0], C_pi, F2x=F2x) + other_rf[0]
        T_j[0,:]      = forcing_to_temperature(T_j_minus1, q[0,:], d, F[0])
    else:
        emissions[0]  = root(infer_emissions, 0., args=(C[0], R_i[0,:],
            tau, a, C_pi))['x']
        F[0]          = co2_log(C[0], C_pi, F2x=F2x) + other_rf[0]
        T_j[0,:]      = forcing_to_temperature(T_j[0,:], q[0,:], d, F[0])

    # Second timestep onwards
    for t in range(1,nt):
        emissions[t], C_acc[t], R_i[t,:], time_scale_sf = (
            inverse_carbon_cycle(
                C[t], C_acc[t-1], np.sum(T_j[t-1,:]), r0, rc, rt, iirf_max, 
                time_scale_sf, a, tau, iirf_h, R_i[t-1,:],
                C_pi, C[t-1], emissions[t-1]
            )
        )
        F[t]     = co2_log(C[t], C_pi, F2x=F2x) + other_rf[t]
        T_j[t,:] = forcing_to_temperature(T_j[t-1,:], q[t,:], d, F[t])

    # Output temperatures
    T = np.sum(T_j, axis=-1)
    if restart_out:
        restart_out_val = (R_i[-1], T_j[-1], C_acc[-1], emissions[-1],
            time_scale_sf, C[-1])
        return emissions, F, T, restart_out_val
    else:
        return emissions, F, T



import numpy as np

"""Default values of thermal response and radiative forcing parameters.

q       : slow and fast coefficients of contribution to temperature
tcr     : transient climate response (K)
ecs     : equilibrium climate sensitivity (K)
d       : slow and fast contributions to global mean temperature change (yr)
f2x     : radiative forcing from a doubling of CO2 (W/m2)
tcr_dbl : timescale for reaching a doubling of CO2 for TCR (yr)
"""

tcr     = 1.6
ecs     = 2.75

q       = np.array([0.33, 0.41])
tcrecs  = np.array([tcr, ecs])
d       = np.array([239.0,4.1])
f2x     = 3.71
tcr_dbl = 69.661 #TODO: v2.0 make explicitly log(2)/log(1.01)


import numpy as np
import os

emissions_filename = os.path.join(os.path.dirname(__file__), 'natural.csv')

class Emissions:
    alldata   = np.loadtxt(emissions_filename, skiprows=4)
    year      = alldata[:,0]
    ch4       = alldata[:,1]
    n2o       = alldata[:,2]
    emissions = alldata[:,1:]



# number of chlorine atoms in each species
CFC11     = 3
CFC12     = 2
CFC113    = 3
CFC114    = 2
CFC115    = 1
CARB_TET  = 4
CCL4      = CARB_TET
MCF       = 3
CH3CCL3   = MCF
HCFC22    = 1
HCFC141B  = 2
HCFC142B  = 1
HALON1211 = 1
HALON1202 = 0
HALON1301 = 0
HALON2402 = 0
CH3BR     = 0
CH3CL     = 1
CH2CL2    = 2
CHCL3     = 3

aslist    = [CFC11, CFC12, CFC113, CFC114, CFC115, CARB_TET, MCF, HCFC22,
             HCFC141B, HCFC142B, HALON1211, HALON1202, HALON1301, HALON2402,
             CH3BR, CH3CL]




from __future__ import division

import numpy as np
from scipy.optimize import root

from ..constants.general import ppm_gtc

"""Carbon cycle function from FaIR v1.0.0."""

def carbon_cycle(e0, c_acc0, temp, r0, rc, rt, iirf_max, time_scale_sf0, a, tau,
    iirf_h, carbon_boxes0, c_pi, c0, e1):
    """Calculates CO2 concentrations from emissions.

    Inputs:
        e0            : emissions of CO2 (GtC) in timestep t-1
        c_acc0        : cumulative airborne carbon anomaly (GtC) since
                        pre-industrial, timestep t-1
        temp          : temperature anomaly above pre-industrial (K)
        r0            : pre-industrial time-integrated airborne fraction (yr)
        rc            : sensitivity of time-integrated airborne fraction to
                        airborne carbon (yr/GtC)
        rt            : sensitivity of time-integrated airborne fraction to
                        temperature (yr/K)
        iirf_max      : maximum value of time-integrated airborne fraction (yr)
        time_scale_sf0: initial guess of alpha scaling factor
        a             : partition coefficient of carbon boxes
        tau           : present-day decay time constants of CO2 (yr)
        iirf_h        : time horizon for time-integrated airborne fraction (yr)
        carbon_boxes0 : carbon stored in each atmospheric reservoir at timestep
                        t-1 (GtC)
        c_pi          : pre-industrial concentration of CO2, ppmv
        c0            : concentration of CO2 in timestep t-1, ppmv
        e1            : emissions of CO2 in timestep t, GtC

    Outputs:
        c1            : concentrations of CO2 in timestep t, ppmv
        c_acc1        : cumulative airborne carbon anomaly (GtC) since
                        pre-industrial, timestep t
        carbon_boxes1 : carbon stored in each atmospheric reservoir at timestep
                        t (GtC)
        time_scale_sf : scale factor for CO2 decay constants
    """
    iirf = _iirf_simple(c_acc0, temp, r0, rc, rt, iirf_max)
    time_scale_sf = root(_iirf_interp, time_scale_sf0,
      args=(a, tau, iirf_h, iirf))['x']
    tau_new = tau * time_scale_sf
    carbon_boxes1 = carbon_boxes0*np.exp(-1.0/tau_new) + a*e1 / ppm_gtc
    c1 = np.sum(carbon_boxes1) + c_pi
    c_acc1 = c_acc0 + 0.5*(e1 + e0) - (c1 - c0)*ppm_gtc
    return c1, c_acc1, carbon_boxes1, time_scale_sf


def _iirf_interp(alp_b,a,tau,iirf_h,targ_iirf):
    """Interpolation function for finding alpha, the CO2 decay time constant
    scaling factor, in iirf_h equation. See Eq. (7) of Millar et al ACP (2017).

    Inputs:
        alp_b    : Guess for alpha, the scale factor, for tau
        a        : partition fractions for CO2 boxes
        tau      : time constants for CO2 boxes
        iirf_h   : time horizon for time-integrated airborne fraction
        targ_iirf: iirf_h calculated using simple parameterisation (Eq. (8),
                   Millar et al (2017)).
    """

    iirf_arr = alp_b*(np.sum(a*tau*(1.0 - np.exp(-iirf_h/(tau*alp_b)))))
    return iirf_arr - targ_iirf


def _iirf_simple(c_acc, temp, r0, rc, rt, iirf_max):
    """Simple linear iIRF relationship. Eq. (8) of Millar et al ACP (2017).

    Inputs:
        c_acc    : cumulative airborne carbon anomaly (GtC) since
                   pre-industrial
        temp     : temperature anomaly since pre-industrial
        r0       : pre-industrial time-integrated airborne fraction (yr)
        rc       : sensitivity of time-integrated airborne fraction to airborne
                   carbon (yr/GtC)
        rt       : sensitivity of time-integrated airborne fraction to
                   temperature (yr/K)
        iirf_max : maximum value of time-integrated airborne fraction (yr)

    Outputs:
        iirf     : time-integrated airborne fraction of carbon (yr)
    """

    return np.min([r0 + rc * c_acc + rt * temp, iirf_max])



# Fractional stratospheric release for ODSs.
# References:
# Daniel, J. and Velders, G.: A focus on information and options for 
#   policymakers, in: Scientific Assessment of Ozone Depletion, WMO, 2011
# Newman et al., 2007: A new formulation of equivalent effective stratospheric
#   chlorine (EESC)

import numpy as np

CFC11     = 0.47
CFC12     = 0.23
CFC113    = 0.29
CFC114    = 0.12
CFC115    = 0.04
CARB_TET  = 0.56
MCF       = 0.67
HCFC22    = 0.13
HCFC141B  = 0.34
HCFC142B  = 0.17
HALON1211 = 0.62
HALON1202 = 0.62
HALON1301 = 0.28
HALON2402 = 0.65
CH3BR     = 0.60
CH3CL     = 0.44
CH2CL2    = np.nan # no literature value available
CHCL3     = np.nan # no literature value available

# This is the list of gases included in the RCPs/AR5/CMIP5.
aslist    = [CFC11, CFC12, CFC113, CFC114, CFC115, CARB_TET, MCF, HCFC22,
             HCFC141B, HCFC142B, HALON1211, HALON1202, HALON1301, HALON2402,
             CH3BR, CH3CL]



from __future__ import division

import numpy as np
from scipy.interpolate import interp1d


def _import_emis_file(rcp):
    if rcp in ['rcp3pd', 'rcp26']:
        from ..RCPs.rcp26 import Emissions as rcp_emis
    elif rcp=='rcp45':
        from ..RCPs.rcp45 import Emissions as rcp_emis
    elif rcp in ['rcp6', 'rcp60']:
        from ..RCPs.rcp60 import Emissions as rcp_emis
    elif rcp=='rcp85':
        from ..RCPs.rcp85 import Emissions as rcp_emis
    else:
        raise ValueError('rcp must be rcp26, rcp45, rcp60 or rcp85')
    return rcp_emis


def scen_open(filename,
              include_cfcs='rcp45',
              startyear=1765,
              harmonise=None):

    """
    Opens a MAGICC6 .SCEN file and extracts the data. Interpolates linearly
    between non-consecutive years in the SCEN file. Fills in chlorinated gases
    from a specified RCP scenario or from custom emissions.

    Inputs:
        filename: the .SCEN file to open

    Keywords:
        include_cfcs: string, False, or nt x 16 numpy array
            MAGICC files do not come loaded with CFCs (indices 24-39).
            They are given in the harmonised files at
            http://www.pik-potsdam.de/~mmalte/rcps/.
            - Specify 'rcp3pd', 'rcp45', 'rcp6' or 'rcp85' to use these RCPs.
            - Use False to ignore and create a 24-species emission file.
            - Provide an array to tack your own chlorinated gases onto the SCEN
        startyear: First year of output file. If before first year of the SCEN
            file, use RCP4.5 to fill
        harmonise: None, or year
            Linearly interpolate between 2000 in the RCP file and the specified
            year. If None, do not harmonise

    Returns:
        nt x 40 numpy emissions array

    nt is defined as <last year of SCEN file> -
        <earlier of startyear and first year of SCEN file> + 1

    It is assumed that the .SCEN files follow the formatting convention on
    the MAGICC wiki at
    http://wiki.magicc.org/index.php?title=Creating_MAGICC_Scenario_Files.
    """

    with open(filename) as f:
        # First line is the number of time steps in the SCEN file
        str_nt = f.readline().strip()
        nt = int(str_nt)

        # Next 6 lines are unused by FaIR.
        for i in range(6):
            f.readline()

        # Eighth line is the column headers. When object orientation is
        # implemented this will be important.
        headers = f.readline().split()

        # Ninth line is the units. Again this will be important in OO-FaIR
        units = f.readline().split()

    # Now the data!
    scen_emissions = np.genfromtxt(filename, skip_header=9, max_rows=nt)
    emissions = np.copy(scen_emissions)
    scen_years = scen_emissions[:,0]

    # Interpolate between non-consecutive years in SCEN file
    f = interp1d(scen_years, emissions[:,1:], kind='linear', axis=0,
        assume_sorted=True, bounds_error=True)
    full_years = np.arange(scen_years[0], scen_years[-1]+1)
    emissions_filled = np.hstack((full_years[:,None], f(full_years)))

    # Add CFCs if requested
    if type(include_cfcs) is np.ndarray:
        if include_cfcs.shape != (len(full_years), 16):
            raise ValueError("If chlorinated gas emissions are provided by " +
            "the user they should be of size nt x 16 where nt is the number "+
            "of consecutive years. Size is %s." % str(include_cfcs.shape))
        emissions_filled = np.append(emissions_filled, include_cfcs, axis=1)
    elif include_cfcs==False:
        pass
    elif include_cfcs.lower()[:3]=='rcp':
        rcp_emis = _import_emis_file(include_cfcs.lower()).emissions
        # Need to ensure only years present in the SCEN file are taken
        # Cheers: https://stackoverflow.com/questions/3522946/
        # using-numpy-arrays-as-lookup-tables
        if int(scen_years[0])<1765:
            raise ValueError("CFCs can only be infilled from RCPs as far "+
            "back as 1765 at present")
        rcp_years = np.arange(scen_years[0], scen_years[-1]+1)
        mapping = dict(zip(rcp_emis[:,0], range(rcp_emis.shape[0])))
        rcp_cfcs = np.array([rcp_emis[mapping[key],24:] for key in rcp_years])
        emissions_filled = np.append(emissions_filled, rcp_cfcs, axis=1)
    else:
        raise ValueError("include_cfcs should be an nt x 16 numpy array, a " +
        "string (rcp3pd, rcp45, rcp6 or rcp85) or False.")

    # Fill in any pre-SCEN years from RCP4.5. All pathways are identical <2000
    if scen_years[0]>startyear:
        if scen_years[0]>2000:
            raise ValueError("Can only fill in history unambiguously if "    +
            "first year in SCEN file is 2000 or earlier. You have requested "+
            "startyear=%d and the first year in SCEN file is %d"
            % (startyear, scen_years[0]))
        else:
            # tack RCP45 on to beginning
            rcp_emis = _import_emis_file('rcp45').emissions
            rcp_years = np.arange(startyear, scen_years[0])
            mapping = dict(zip(rcp_emis[:,0], range(rcp_emis.shape[0])))
            if include_cfcs==False:
                rcp_all = np.array([rcp_emis[mapping[key],:24] for key in rcp_years])
            else:
                rcp_all = np.array([rcp_emis[mapping[key],:] for key in rcp_years])
            emissions_filled = np.insert(emissions_filled, 0, rcp_all, axis=0)

        # harmonise?
        if harmonise is not None:
            harmonise = int(harmonise)
            if harmonise < 2000:
                raise ValueError("Cannot harmonise before 2000.")
            elif harmonise > scen_years[-1]:
                 raise ValueError("Cannot harmonise after last year of "      +
                 "input dataset")
            rcp_emis_2000 = rcp_emis[2000-startyear,:]
            for j in range(1,emissions_filled.shape[1]):
                f = interp1d((2000,harmonise), (rcp_emis_2000[j],
                    emissions_filled[harmonise-startyear,j]))
                emissions_filled[2000-startyear:harmonise-startyear,j] = f(
                    np.arange(2000,harmonise))

    return emissions_filled



from __future__ import division

import inspect
import numpy as np
import warnings

from .ancil import natural, cmip6_volcanic, cmip6_solar, historical_scaling
from .constants import molwt, lifetime, radeff
from .constants.general import M_ATMOS, ppm_gtc
from .defaults import carbon, thermal
from .forcing import ozone_tr, ozone_st, h2o_st, contrails, aerosols, bc_snow,\
                                         landuse
from .gas_cycle.gir import calculate_alpha, step_concentration
from .gas_cycle.fair1 import carbon_cycle
from .forcing.ghg import co2_log, minor_gases
from .temperature.millar import calculate_q



# TODO: unified interface to the different carbon cycles


def emis_to_conc(c0, e0, e1, ts, lt, vm):
    """Calculate concentrations of well mixed GHGs from emissions for simple
    one-box model.
    
    Inputs (all can be scalar or 1D arrays for multiple species):
        c0: concentrations in timestep t-1
        e0: emissions in timestep t-1
        e1: emissions in timestep t
        ts: length of timestep. Use 1 for sensible results in FaIR 1.3.
        lt: atmospheric (e-folding) lifetime of GHG
        vm: conversion from emissions units (e.g. Mt) to concentrations units
            (e.g. ppb)
            
    Outputs:
        c1: concentrations in timestep t
    """
    c1 = c0 - c0 * (1.0 - np.exp(-ts/lt)) + 0.5 * ts * (e1 + e0) * vm
    return c1


def fair_scm(
    emissions=False,
    emissions_driven=True,
    C=None,
    other_rf=0.0,
    q        = thermal.q,
    tcrecs   = thermal.tcrecs,
    d        = thermal.d,
    F2x      = thermal.f2x,
    tcr_dbl  = thermal.tcr_dbl,
    a        = carbon.a,
    tau      = carbon.tau,
    r0       = carbon.r0,
    rc       = carbon.rc,
    rt       = carbon.rt,
    iirf_max = carbon.iirf_max,
    iirf_h   = carbon.iirf_h,
    C_pi=np.array([278., 722., 273., 34.497] + [0.]*25 + [13.0975, 547.996]),
    E_pi=np.zeros(40),
    restart_in=False,
    restart_out=False,
    F_tropO3 = 0.,
    F_aerosol = 0.,
    F_volcanic=cmip6_volcanic.Forcing.volcanic,
    F_solar=cmip6_solar.Forcing.solar,
    F_contrails=0.,
    F_bcsnow=0.,
    F_landuse=0.,
    aviNOx_frac=0.,
    F_ref_aviNOx=0.0448,
    E_ref_aviNOx=2.946,
    F_ref_BC=0.04,
    E_ref_BC=8.09,
    fossilCH4_frac=0.,
    natural=natural.Emissions.emissions,
    efficacy=np.array([1.]*9 + [3.] + [1.]*3),
    scale=None,
    oxCH4_frac=0.61,
    ghg_forcing="Etminan",
    scale_F2x=True,
    stwv_from_ch4=None,
    b_aero = np.array([-6.2227e-3, 0.0, -3.8392e-4, -1.16551e-3, 1.601537e-2,
      -1.45339e-3, -1.55605e-3]),
    b_tro3 = np.array([2.8249e-4, 1.0695e-4, -9.3604e-4, 99.7831e-4]),
    pi_tro3 =np.array([722, 170, 10, 4.29]),
    ghan_params = np.array([-1.95011431, 0.01107147, 0.01387492]),
    stevens_params = np.array([0.001875, 0.634, 60.]),
    ref_isSO2=True, # is Stevens SO2 emissions in units SO2 (T) or S (F)
    useMultigas=True,
    useStevenson=True,   # deprecate this switch in v1.7
    tropO3_forcing='stevenson',
    lifetimes=False,
    aerosol_forcing="aerocom+ghan",
    scaleAerosolAR5=True,
    fixPre1850RCP=True,
    useTropO3TFeedback=True,
    scaleHistoricalAR5=False,
    contrail_forcing='NOx',
    kerosene_supply=0.,
    landuse_forcing='co2',
    aCO2land=-0.00113789,
    ariaci_out=False,
    bcsnow_forcing='emissions',
    diagnostics=None,
    gir_carbon_cycle=False,
    temperature_function='Millar',
    lambda_global=1.18,  # this and the below only used in two-layer model
    ocean_heat_capacity=np.array([8.2, 109.0]),
    ocean_heat_exchange=0.67,
    deep_ocean_efficacy=1.28,
    ):

    # Prevents later errors when SLCFs not specified
    if type(emissions) is bool and not emissions_driven:
        tropO3_forcing='external'

    if useStevenson is not None:
        warnings.warn('"useStevenson" will be deprecated in the future; use '+
          'tropO3_forcing keyword with "cmip6", "stevenson", "regression" or "external"',
          DeprecationWarning)

    # is iirf_h < iirf_max? Don't stop the code, but warn user
    if iirf_h < iirf_max:
        warnings.warn('iirf_h=%f, which is less than iirf_max (%f)'
          % (iirf_h, iirf_max), RuntimeWarning)

    # Conversion between ppb/ppt concentrations and Mt/kt emissions
    # in the RCP databases ppb = Mt and ppt = kt so factor always 1e18
    emis2conc = M_ATMOS/1e18*np.asarray(molwt.aslist)/molwt.AIR

    # Funny units for nitrogen emissions - N2O is expressed in N2 equivalent
    n2o_sf = molwt.N2O/molwt.N2
    emis2conc[2] = emis2conc[2] / n2o_sf

    # Convert any list to a numpy array for (a) speed and (b) consistency.
    # Goes through all variables in scope and converts them.
    frame = inspect.currentframe()
    args, _, _, values = inspect.getargvalues(frame)
    for arg_to_check in args:
        if type(values[arg_to_check]) is list:
            exec(arg_to_check + '= np.array(' + arg_to_check + ')')

    # Initialise simplified carbon cycle parameters
    if gir_carbon_cycle:
        g1 = np.sum(a*tau * (1 - (1 + iirf_h/tau) * np.exp(-iirf_h/tau)))
        g0 = 1/(np.sinh(np.sum(a*tau*(1 - np.exp(-iirf_h/tau)) , axis=-1)/g1))
        if useMultigas:
            cumulative_emissions = np.cumsum(emissions[:,1:3].sum(axis=1))
        else:
            cumulative_emissions = np.cumsum(emissions)
        airborne_emissions = np.zeros_like(cumulative_emissions)

    # import correct conversion
    if temperature_function=='Millar':
        from .temperature.millar import forcing_to_temperature
    elif temperature_function=='Geoffroy':
        from .temperature.geoffroy import forcing_to_temperature
    else:
        raise ValueError('temperature_function must be "Millar" or "Geoffroy"')
        
    # Set up the output timeseries variables depending on options and perform
    # basic sense checks
    if useMultigas:
        ngas = 31
        if diagnostics=='AR6':
            iF_tro3 = 31
            iF_sto3 = 32
            iF_ch4h = 33
            iF_cont = 34
            iF_aerd = 35
            iF_aeri = 36
            iF_bcsn = 37
            iF_luch = 38
            iF_volc = 39
            iF_solr = 40
            nF = 41
        else:
            iF_tro3 = 4
            iF_sto3 = 5
            iF_ch4h = 6
            iF_cont = 7
            iF_aer  = 8
            iF_bcsn = 9
            iF_luch = 10
            iF_volc = 11
            iF_solr = 12
            nF = 13
        if emissions_driven:
            if type(emissions) is not np.ndarray or emissions.shape[1] != 40:
                raise ValueError(
                  "emissions timeseries should be a nt x 40 numpy array")
            carbon_boxes_shape = (emissions.shape[0], a.shape[0])
            if temperature_function=='Millar':
                thermal_boxes_shape = (emissions.shape[0], d.shape[0])
            else:
                thermal_boxes_shape = (emissions.shape[0], d.shape[0], 2)
            nt = emissions.shape[0]
        else:
            if type(C) is not np.ndarray or C.shape[1] != ngas:
                raise ValueError(
                  "C timeseries should be a nt x %d numpy array" % ngas)
            if temperature_function=='Millar':
                thermal_boxes_shape = (C.shape[0], d.shape[0])
            else:
                thermal_boxes_shape = (C.shape[0], d.shape[0], 2)
            nt = C.shape[0]
        if np.isscalar(fossilCH4_frac):
            fossilCH4_frac = np.ones(nt) * fossilCH4_frac
        # If custom gas lifetimes are supplied, use them, else import defaults
        if type(lifetimes) is np.ndarray:
            if len(lifetimes)!=ngas:
                raise ValueError(
                  "custom GHG lifetime array must have " + str(ngas) + 
                  " elements")
        else:
            lifetimes = lifetime.aslist
        # Select the desired GHG forcing relationship and populate 
        # stratospheric water vapour from methane scale factor if not specified
        # by user
        if ghg_forcing.lower()=="etminan":
            from .forcing.ghg import etminan as ghg
            if stwv_from_ch4==None: stwv_from_ch4=0.12
        elif ghg_forcing.lower()=="myhre":
            from .forcing.ghg import myhre as ghg
            if stwv_from_ch4==None: stwv_from_ch4=0.15
        else:
            raise ValueError(
              "ghg_forcing should be 'etminan' (default) or 'myhre'")
        # aerosol breakdown
        ariaci = np.zeros((nt,2))
            
        # Check natural emissions and convert to 2D array if necessary
        if emissions_driven: # don't check for conc runs
            if type(natural) in [float,int]:
                natural = natural * np.ones((nt,2))
            elif type(natural) is np.ndarray:
                if natural.ndim==1:
                    if natural.shape[0]!=2:
                        raise ValueError(
                          "natural emissions should be a 2-element or nt x 2 " +
                          "array")
                    natural = np.tile(natural, nt).reshape((nt,2))
                elif natural.ndim==2:
                    if natural.shape[1]!=2 or natural.shape[0]!=nt:
                        raise ValueError(
                          "natural emissions should be a 2-element or nt x 2 " +
                          "array")
            else:
                raise ValueError(
                  "natural emissions should be a scalar, 2-element, or nt x 2 " +
                  "array")

        # check scale factor is correct shape. If 1D inflate to 2D
        if scale is None:
            scale = np.ones((nt,nF))
        elif scale.shape[-1]==nF:
            if scale.ndim==2 and scale.shape[0]==nt:
                pass
            elif scale.ndim==1:
                scale = np.tile(scale, nt).reshape((nt,nF))
        else:
            raise ValueError("in multi-gas mode, scale should be None, or a "+
              "(%d,) or (%d, 13) array" % (nF, nF))

        # if scaling the historical time series to match AR5, apply these
        # factors to whatever the user specifies
        if scaleHistoricalAR5:
            scale=scale*historical_scaling.all[:nt,:]

        # if tropospheric ozone is directly specified and scalar, inflate to
        # 1D array. Raise ValueError if wrong shape
        if tropO3_forcing[0].lower()=='e':
            if type(F_tropO3) is np.ndarray:
                if F_tropO3.shape[0]!=nt or F_tropO3.ndim!=1:
                    raise ValueError("F_tropO3 should be a scalar or (nt,) "+
                    "array")
            elif type(F_tropO3) in [float,int]:
                F_tropO3 = F_tropO3 * np.ones(nt)
            else:
                raise ValueError("F_tropO3 should be a scalar or (nt,) array")

    else:
        ngas = 1
        nF   = 1

        if emissions_driven:
            if type(emissions) is np.ndarray:
                if emissions.ndim != 1:
                    raise ValueError(
                      "In CO2-only mode, emissions should be a 1D array")
                nt = emissions.shape[0]
                carbon_boxes_shape = (nt, a.shape[0])
                if temperature_function=='Millar':
                    thermal_boxes_shape = (nt, d.shape[0])
                else:
                    thermal_boxes_shape = (nt, d.shape[0], 2)
            elif type(other_rf) is np.ndarray:
                if other_rf.ndim != 1:
                    raise ValueError(
                      "In CO2-only mode, other_rf should be a 1D array")
                nt = other_rf.shape[0]
                carbon_boxes_shape = (nt, a.shape[0])
                if temperature_function=='Millar':
                    thermal_boxes_shape = (nt, d.shape[0])
                else:
                    thermal_boxes_shape = (nt, d.shape[0], 2)
                emissions = np.zeros(nt)
            else:
                raise ValueError(
                  "Neither emissions or other_rf is defined as a timeseries")

        else:
            if type(C) is not np.ndarray or C.ndim != 1:
                raise ValueError(
                  "In CO2-only mode, concentrations should be a 1D array")
            nt = C.shape[0]
            if temperature_function=='Millar':
                thermal_boxes_shape = (nt, d.shape[0])
            else:
                thermal_boxes_shape = (nt, d.shape[0], 2)
            # expand C to 2D array for consistency with other calcs
            C = C.reshape((nt, 1))

        # check scale factor is correct shape - either scalar or 1D
        # needs try/except really
        if scale is None:
            scale = np.ones(nt)
        elif np.isscalar(scale):
            scale = np.ones(nt) * scale
        elif scale.ndim==1 and scale.shape[0]==nt:
            pass
        else:
            raise ValueError("in CO2-only mode, scale should be None, a "+
              "scalar or a (nt,) array")

        # if scaling the historical time series to match AR5, apply these
        # factors to whatever the user specifies
        if scaleHistoricalAR5:
            scale=scale*historical_scaling.co2[:nt]

    # If TCR and ECS are supplied, calculate q coefficients
    if type(tcrecs) is np.ndarray and temperature_function=='Millar':
        q = calculate_q(tcrecs, d, F2x, tcr_dbl, nt)

    # Check a and tau are same size
    if a.ndim != 1:
        raise ValueError("a should be a 1D array")
    if tau.ndim != 1:
        raise ValueError("tau should be a 1D array")
    if len(a) != len(tau):
        raise ValueError("a and tau should be the same size")
    if not np.isclose(np.sum(a), 1.0):
        raise ValueError("a should sum to one")

    # Allocate intermediate and output arrays
    F = np.zeros((nt, nF))
    C_acc = np.zeros(nt)
    T_j = np.zeros(thermal_boxes_shape)
    T = np.zeros(nt)
    C_0 = np.copy(C_pi)
    if emissions_driven:
        C = np.zeros((nt, ngas))
        R_i = np.zeros(carbon_boxes_shape)

    if temperature_function!='Millar':
        heatflux = np.zeros(nt)
        ohc = np.zeros(nt)
        lambda_eff = np.zeros(nt)

    if restart_in:
        R_minus1 = restart_in[0]
        T_j_minus1 = restart_in[1]
        C_acc_minus1 = restart_in[2]
        E_minus1 = restart_in[3]
        C_minus1 = np.sum(R_minus1,axis=-1) + C_0[0]

        if gir_carbon_cycle:
            raise NotImplementedError('GIR carbon cycle not configured to ' +
                'work with restarts')
        else:
            C[0,0], C_acc[0], R_i[0,:], time_scale_sf = carbon_cycle(
              E_minus1,
              C_acc_minus1,
              np.sum(T_j_minus1),
              r0,
              rc,
              rt,
              iirf_max,
              0.16,
              a,
              tau,
              iirf_h,
              R_minus1,
              C_pi[0],
              C_minus1,
              emissions[0]
            )

        if np.isscalar(other_rf):
            F[0,0] = co2_log(C[0,0], C_pi[0], F2x) + other_rf
        else:
            F[0,0] = co2_log(C[0,0], C_pi[0], F2x) + other_rf[0]

        F[0,0] = F[0,0] * scale[0]

        if temperature_function=='Millar':
            T_j[0,:] = forcing_to_temperature(T_j_minus1, q[0,:], d, F[0,:])
            T[0]=np.sum(T_j[0,:])
        else:
            # leave unimplemented unless somebody invents a use case
            raise(NotImplementedError('Restarts not implemented with Geoffroy '+
                'temperature function'))

    else:
        # Initialise the carbon pools to be correct for first timestep in
        # numerical method
        if emissions_driven:
            if useMultigas:
                R_i[0,:] = a * (np.sum(emissions[0,1:3])) / ppm_gtc
                C[0,1:] = C_0[1:]
            else:
                R_i[0,:] = a * emissions[0,np.newaxis] / ppm_gtc
            C[0,0] = np.sum(R_i[0,:],axis=-1) + C_0[0]

    if useMultigas:
        # CO2, CH4 and N2O are co-dependent
        F[0,0:3] = ghg(C[0,0:3], C_pi[0:3], F2x=F2x, scale_F2x=scale_F2x)
        # Minor (F- and H-gases) are linear in concentration
        # the factor of 0.001 here is because radiative efficiencies are given
        # in W/m2/ppb and concentrations of minor gases are in ppt.
        if diagnostics=='AR6':
            F[0,3:31] = minor_gases(C[0,3:], C_pi[3:])
        else:
            F[0,3] = np.sum(minor_gases(C[0,3:], C_pi[3:]))

        # Tropospheric ozone:
        # v1.5 update: don't require emissions driven runs here for GHGs
        # because SLCFs can still be given as emissions with GHGs as
        # concentrations
        if type(emissions) is not bool:
            if useStevenson and tropO3_forcing[0].lower()=='s':
                F[0,iF_tro3] = ozone_tr.stevenson(emissions[0,:], C[0,1],
                  T=np.sum(T_j[0,:]), 
                  feedback=useTropO3TFeedback,
                  fix_pre1850_RCP=fixPre1850RCP,
                  PI=pi_tro3)
            elif tropO3_forcing[0].lower()=='c':
                F[0,iF_tro3] = ozone_tr.cmip6_stevenson(emissions[0,:], C[0,1],
                  T=np.sum(T_j[0,:]),
                  feedback=useTropO3TFeedback,
                  PI=np.array([C_pi[1],E_pi[6],E_pi[7],E_pi[8]]),
                  beta=b_tro3)
            elif not useStevenson or tropO3_forcing[0].lower()=='r':
                F[0,iF_tro3] = ozone_tr.regress(emissions[0,:]-E_pi[:], beta=b_tro3)
            else:
                F[0,iF_tro3] = F_tropO3[0]
        else:
            F[0,iF_tro3] = F_tropO3[0]

        # Stratospheric ozone depends on concentrations of ODSs (index 15-30)
        F[0,iF_sto3] = ozone_st.magicc(C[0,15:], C_pi[15:])

        # Stratospheric water vapour is a function of the methane ERF
        F[0,iF_ch4h] = h2o_st.linear(F[0,1], ratio=stwv_from_ch4)

        # Forcing from contrails. No climate feedback so can live outside
        # of forward model in this version
        # v1.5 update: don't require emissions driven runs here for GHGs
        # because SLCFs can still be given as emissions with GHGs as
        # concentrations
        if type(emissions) is not bool:
            if contrail_forcing.lower()[0]=='n':   # from NOx emissions
                F[:,iF_cont] = contrails.from_aviNOx(emissions, aviNOx_frac,
                  F_ref=F_ref_aviNOx, E_ref=E_ref_aviNOx)
            elif contrail_forcing.lower()[0]=='f': # from kerosene production
                F[:,iF_cont] = contrails.from_fuel(kerosene_supply)
            elif contrail_forcing.lower()[0]=='e': # external forcing timeseries
                F[:,iF_cont] = F_contrails
            else:
                raise ValueError("contrails must be one of 'NOx' (estimated "+
                 "from NOx emissions), 'fuel' (estimated from annual jet fuel "+
                 "supplied) or 'external' (an external forcing time series).")
        else:
            F[:,iF_cont] = F_contrails

        # Forcing from aerosols - again no feedback dependence
        # v1.5 update: don't require emissions driven runs here for GHGs
        # because SLCFs can still be given as emissions with GHGs as
        # concentrations

        # TODO: this whole code block is a mess!
        if type(emissions) is not bool:
            if aerosol_forcing.lower()=='stevens':
                ariaci[:,0], ariaci[:,1] = aerosols.Stevens(
                  emissions, stevens_params=stevens_params, E_pi=E_pi[5], 
                  ref_isSO2=ref_isSO2)
                if diagnostics=='AR6':
                    F[:,iF_aerd] = ariaci[:,0]
                    F[:,iF_aeri] = ariaci[:,1]
                else:
                    F[:,iF_aer] = np.sum(ariaci, axis=1)
            elif 'aerocom' in aerosol_forcing.lower():
                ariaci[:,0] = aerosols.aerocom_direct(emissions, beta=b_aero,
                  E_pi=E_pi)
                if 'ghan2' in aerosol_forcing.lower():
                    ariaci[:,1] = aerosols.ghan2(emissions, E_pi, ghan_params)
                elif 'ghan' in aerosol_forcing.lower():
                    ariaci[:,1] = aerosols.ghan_indirect(emissions,
                      scale_AR5=scaleAerosolAR5,
                      fix_pre1850_RCP=fixPre1850RCP,
                      ghan_params=ghan_params,
                      E_pi=E_pi)
                elif 'stevens' in aerosol_forcing.lower():
                    _, ariaci[:,1] = aerosols.Stevens(
                      emissions, stevens_params=stevens_params, E_pi=E_pi[5],
                      ref_isSO2=ref_isSO2)
                if diagnostics=='AR6':
                    F[:,iF_aerd] = ariaci[:,0]
                    F[:,iF_aeri] = ariaci[:,1]
                else:
                    F[:,iF_aer] = np.sum(ariaci, axis=1)
            elif aerosol_forcing.lower()[0] == 'e':
                if diagnostics!='AR6':
                    F[:,iF_aer] = F_aerosol
                    ariaci[:] = np.nan
                else:
                    raise ValueError('AR6 diagnostics not compatible with ' +
                'externally forced aerosols')
            else:
                raise ValueError("aerosol_forcing should be one of 'stevens', " +
                  "aerocom, aerocom+ghan, aerocom+stevens or external")
        else:
            if diagnostics!='AR6':
                F[:,iF_aer] = F_aerosol
                ariaci[:] = np.nan
            else:
                raise ValueError('AR6 diagnostics not compatible with ' +
            'externally forced aerosols')

        # Black carbon on snow - no feedback dependence
        # v1.5 update: don't require emissions driven runs here for GHGs
        # because SLCFs can still be given as emissions with GHGs as
        # concentrations
        if type(emissions) is not bool:
           if bcsnow_forcing.lower()[0]=='e':
               F[:,iF_bcsn] = bc_snow.linear(emissions-E_pi, F_ref=F_ref_BC,
                   E_ref=E_ref_BC)
           else:
               F[:,iF_bcsn] = F_bcsnow
        else:
            F[:,iF_bcsn] = F_bcsnow

        # Land use change - either use a scaling with cumulative CO2 emissions
        # or an external time series
        # v1.5 update: don't require emissions driven runs here for GHGs
        # because SLCFs can still be given as emissions with GHGs as
        # concentrations
        if type(emissions) is not bool:
            if landuse_forcing.lower()[0]=='c':
                F[:,iF_luch] = landuse.cumulative(emissions-E_pi, aCO2land=aCO2land)
            elif landuse_forcing.lower()[0]=='e':
                F[:,iF_luch] = F_landuse
            else:
                raise ValueError(
                "landuse_forcing should be one of 'co2' or 'external'")
        else:
            F[:,iF_luch] = F_landuse

        # Volcanic and solar copied straight to the output arrays
        F[:,iF_volc] = F_volcanic
        F[:,iF_solr] = F_solar

        # multiply by scale factors
        F[0,:] = F[0,:] * scale[0,:]

    else:
        if np.isscalar(other_rf):
            F[0,0] = co2_log(C[0,0], C_pi[0], F2x) + other_rf
        else:
            F[0,0] = co2_log(C[0,0], C_pi[0], F2x) + other_rf[0]
        F[0,0] = F[0,0] * scale[0]

    if restart_in == False:
        # Update the thermal response boxes
        if temperature_function=='Millar':
            T_j[0,:] = (q[0,:]/d)*(np.sum(F[0,:]))
            T[0] = np.sum(T_j[0,:])
        else:
            T_j[0,:,:], heatflux[0], del_ohc, lambda_eff[0] = forcing_to_temperature(
                T_j[0,:,:],
                np.sum(F[0,:]),
                np.sum(F[0,:]),
                lambda_global=lambda_global,
                ocean_heat_capacity=ocean_heat_capacity,
                ocean_heat_exchange=ocean_heat_exchange,
                deep_ocean_efficacy=deep_ocean_efficacy,
                dt=1
            )
            T[0] = np.sum(T_j[0,:,:], axis=1)[0]
            ohc[0] = ohc[0] + del_ohc


    for t in range(1,nt):

        if emissions_driven:
            if useMultigas:
                if t == 1:
                    time_scale_sf = 0.16
                # Calculate concentrations
                # a. CARBON DIOXIDE
                # Firstly add any oxidised methane from last year to the CO2
                # pool
                oxidised_CH4 = ((C[t-1,1]-C_pi[1]) *
                  (1.0 - np.exp(-1.0/lifetimes[1])) * 
                  (molwt.C/molwt.CH4 * 0.001 * oxCH4_frac * fossilCH4_frac[t]))
                oxidised_CH4 = np.max((oxidised_CH4, 0))

                if gir_carbon_cycle:
                    time_scale_sf = calculate_alpha(
                        cumulative_emissions[t-1],
                        airborne_emissions[t-1],
                        T[t-1],
                        r0, rc, rt, g0, g1)
                    C[t,0], R_i[t,:], airborne_emissions[t] = step_concentration(
                        R_i[t-1,:] + oxidised_CH4,
                        np.sum(emissions[t-1,1:3]),
                        time_scale_sf,
                        a,
                        tau,
                        C_pi[0],
                    )
                else:
                    C[t,0], C_acc[t], R_i[t,:], time_scale_sf = carbon_cycle(
                      np.sum(emissions[t-1,1:3]),
                      C_acc[t-1],
                      T[t-1],
                      r0,
                      rc,
                      rt,
                      iirf_max,
                      time_scale_sf,
                      a,
                      tau,
                      iirf_h,
                      R_i[t-1,:] + oxidised_CH4,
                      C_pi[0],
                      C[t-1,0],
                      np.sum(emissions[t,1:3])
                    )

                # b. METHANE
                C[t,1] = emis_to_conc(
                    C[t-1,1],
                    emissions[t-1,3]+natural[t,0], 
                    emissions[t,3]+natural[t,0],
                    1.0,
                    lifetimes[1],
                    1.0/emis2conc[1]
                    )

                # c. NITROUS OXIDE
                C[t,2] = emis_to_conc(
                    C[t-1,2],
                    emissions[t-1,4]+natural[t,1], 
                    emissions[t,4]+natural[t,1],
                    1.0,
                    lifetimes[2],
                    1.0/emis2conc[2]
                    )

                # d. OTHER WMGHGs
                C[t,3:] = emis_to_conc(
                    C[t-1,3:],
                    emissions[t-1,12:], 
                    emissions[t,12:],
                    1.0,
                    np.array(lifetimes[3:]),
                    1.0/emis2conc[3:]
                    )

                # 2. Radiative forcing
                F[t,0:3] = ghg(C[t,0:3], C_pi[0:3], F2x=F2x, scale_F2x=scale_F2x)
                if diagnostics=='AR6':
                    F[t,3:31] = minor_gases(C[t,3:], C_pi[3:])
                else:
                    F[t,3] = np.sum(minor_gases(C[t,3:], C_pi[3:]))

                if useStevenson and tropO3_forcing[0].lower()=='s':
                    F[t,iF_tro3] = ozone_tr.stevenson(emissions[t,:],
                      C[t,1],
                      T=T[t-1], 
                      feedback=useTropO3TFeedback,
                      fix_pre1850_RCP=fixPre1850RCP,
                      PI=pi_tro3)
                elif tropO3_forcing[0].lower()=='c':
                    F[t,iF_tro3] = ozone_tr.cmip6_stevenson(emissions[t,:], C[t,1],
                      T=np.sum(T_j[t,:]),
                      feedback=useTropO3TFeedback,
                      PI=np.array([C_pi[1],E_pi[6],E_pi[7],E_pi[8]]),
                      beta=b_tro3)
                elif not useStevenson or tropO3_forcing[0].lower()=='r':
                    F[t,iF_tro3] = ozone_tr.regress(emissions[t,:]-E_pi, beta=b_tro3)
                else:
                    F[t,iF_tro3] = F_tropO3[t]
                F[t,iF_sto3] = ozone_st.magicc(C[t,15:], C_pi[15:])
                F[t,iF_ch4h] = h2o_st.linear(F[t,1], ratio=stwv_from_ch4)

                # multiply by scale factors
                F[t,:] = F[t,:] * scale[t,:]

                # 3. Temperature
                # Update the thermal response boxes
                if temperature_function=='Millar':
                    T_j[t,:] = forcing_to_temperature(
                      T_j[t-1,:], q[t,:], d, F[t,:], e=efficacy)
                    T[t] = np.sum(T_j[t,:])
                else:
                    T_j[t,:,:], heatflux[t], del_ohc, lambda_eff[t] = forcing_to_temperature(
                        T_j[t-1,:,:],
                        np.sum(F[t-1,:]),
                        np.sum(F[t,:]),
                        lambda_global=lambda_global,
                        ocean_heat_capacity=ocean_heat_capacity,
                        ocean_heat_exchange=ocean_heat_exchange,
                        deep_ocean_efficacy=deep_ocean_efficacy,
                        dt=1
                    )
                    T[t] = np.sum(T_j[t,:,:], axis=1)[0]
                    ohc[t] = ohc[t-1] + del_ohc

            else:
                if t == 1:
                    time_scale_sf = 0.16
                if gir_carbon_cycle:
                    time_scale_sf = calculate_alpha(
                        cumulative_emissions[t-1],
                        airborne_emissions[t-1],
                        T[t-1],
                        r0, rc, rt, g0, g1)
                    C[t,0], R_i[t,:], airborne_emissions[t] = step_concentration(
                        R_i[t-1,:] + oxidised_CH4,
                        emissions[t-1],
                        time_scale_sf,
                        a,
                        tau,
                        C_pi[0],
                    )
                else:
                    C[t,0], C_acc[t], R_i[t,:], time_scale_sf = carbon_cycle(
                      emissions[t-1],
                      C_acc[t-1],
                      T[t-1],
                      r0,
                      rc,
                      rt,
                      iirf_max,
                      time_scale_sf,
                      a,
                      tau,
                      iirf_h,
                      R_i[t-1,:],
                      C_pi[0],
                      C[t-1,0],
                      emissions[t]
                    )

                if np.isscalar(other_rf):
                    F[t,0] = co2_log(C[t,0], C_pi[0], F2x) + other_rf
                else:
                    F[t,0] = co2_log(C[t,0], C_pi[0], F2x) + other_rf[t]

                F[t,0] = F[t,0] * scale[t]

                if temperature_function=='Millar':
                    T_j[t,:] = forcing_to_temperature(
                      T_j[t-1,:], q[t,:], d, F[t,:])
                    T[t] = np.sum(T_j[t,:])
                else:
                    T_j[t,:,:], heatflux[t], del_ohc, lambda_eff[t] = forcing_to_temperature(
                        T_j[t-1,:,:],
                        np.sum(F[t-1,:]),
                        np.sum(F[t,:]),
                        lambda_global=lambda_global,
                        ocean_heat_capacity=ocean_heat_capacity,
                        ocean_heat_exchange=ocean_heat_exchange,
                        deep_ocean_efficacy=deep_ocean_efficacy,
                        dt=1
                    )
                    T[t] = np.sum(T_j[t,:,:], axis=1)[0]
                    ohc[t] = ohc[t-1] + del_ohc


        else:

            if useMultigas:
                F[t,0:3] = ghg(C[t,0:3], C_pi[0:3], F2x=F2x)
                F[t,3] = np.sum((C[t,3:] - C_pi[3:]) * radeff.aslist[3:]
                  * 0.001)
                if type(emissions) is not bool:
                    if useStevenson and tropO3_forcing[0].lower()=='s':
                        F[t,4] = ozone_tr.stevenson(emissions[t,:]-E_pi,
                          C[t,1],
                          T=T[t-1],
                          feedback=useTropO3TFeedback,
                          fix_pre1850_RCP=fixPre1850RCP)
                    elif tropO3_forcing[0].lower()=='c':
                        F[t,4] = ozone_tr.cmip6_stevenson(emissions[t,:], C[t,1],
                          T=np.sum(T_j[t,:]),
                          feedback=useTropO3TFeedback,
                          PI=np.array([C_pi[1],E_pi[6],E_pi[7],E_pi[8]]),
                          beta=b_tro3)
                    elif not useStevenson or tropO3_forcing[0].lower()=='r':
                        F[t,4] = ozone_tr.regress(emissions[t,:]-E_pi, beta=b_tro3)
                    else:
                        F[t,4] = F_tropO3[t]
                else:
                    F[t,4] = F_tropO3[t]
                F[t,5] = ozone_st.magicc(C[t,15:], C_pi[15:])
                F[t,6] = h2o_st.linear(F[t,1], ratio=stwv_from_ch4)

                # multiply by scale factors
                F[t,:] = F[t,:] * scale[t,:]

                # 3. Temperature
                # Update the thermal response boxes
                if temperature_function=='Millar':
                    T_j[t,:] = forcing_to_temperature(
                      T_j[t-1,:], q[t,:], d, F[t,:], e=efficacy)
                    T[t] = np.sum(T_j[t,:])
                else:
                    T_j[t,:,:], heatflux[t], del_ohc, lambda_eff[t] = forcing_to_temperature(
                        T_j[t-1,:,:],
                        np.sum(F[t-1,:]),
                        np.sum(F[t,:]),
                        lambda_global=lambda_global,
                        ocean_heat_capacity=ocean_heat_capacity,
                        ocean_heat_exchange=ocean_heat_exchange,
                        deep_ocean_efficacy=deep_ocean_efficacy,
                        dt=1
                    )
                    T[t] = np.sum(T_j[t,:,:], axis=1)[0]
                    ohc[t] = ohc[t-1] + del_ohc

            else:
                if np.isscalar(other_rf):
                    F[t,0] = co2_log(C[t,0], C_pi[0], F2x) + other_rf
                else:
                    F[t,0] = co2_log(C[t,0], C_pi[0], F2x) + other_rf[t]

                F[t,0] = F[t,0] * scale[t]

                if temperature_function=='Millar':
                    T_j[t,:] = forcing_to_temperature(T_j[t-1,:], q[t,:], d, F[t,:])
                    T[t] = np.sum(T_j[t,:])
                else:
                    T_j[t,:,:], heatflux[t], del_ohc, lambda_eff[t] = forcing_to_temperature(
                        T_j[t-1,:,:],
                        np.sum(F[t-1], axis=1),
                        np.sum(F[t], axis=1),
                        lambda_global=lambda_global,
                        ocean_heat_capacity=ocean_heat_capacity,
                        ocean_heat_exchange=ocean_heat_exchange,
                        deep_ocean_efficacy=deep_ocean_efficacy,
                        dt=1
                    )
                    T[t] = np.sum(T_j[t,:,:], axis=1)[0]
                    ohc[t] = ohc[t-1] + del_ohc

    if not useMultigas:
        C = np.squeeze(C)
        F = np.squeeze(F)

    if restart_out:
        if useMultigas:
            E_minus1 = np.sum(emissions[-1,1:3])
        else:
            E_minus1 = emissions[-1]
        restart_out_val=(R_i[-1],T_j[-1],C_acc[-1],E_minus1)
        return C, F, T, restart_out_val

    if ariaci_out:
        if temperature_function=='Geoffroy':
            return C, F, T, ariaci, lambda_eff, ohc, heatflux
        else:
            return C, F, T, ariaci
    else:
        if temperature_function=='Geoffroy':
            if gir_carbon_cycle:
                return C, F, T, lambda_eff, ohc, heatflux, airborne_emissions/cumulative_emissions
            else:
                return C, F, T, lambda_eff, ohc, heatflux
        else:
            return C, F, T




import numpy as np
import os

forcing_filename = os.path.join(os.path.dirname(__file__),
    'cmip5_annex2_forcing.csv')

class Forcing:
    forcing  = np.loadtxt(forcing_filename, skiprows=1, delimiter=',')
    year     = forcing[:,0]
    co2      = forcing[:,1]
    ghg_other= forcing[:,2]
    tropo3   = forcing[:,3]
    strato3  = forcing[:,4]
    aero     = forcing[:,5]
    landuse  = forcing[:,6]
    stwv     = forcing[:,7]
    bcsnow   = forcing[:,8]
    contrails= forcing[:,9]
    solar    = forcing[:,10]
    volcanic = forcing[:,11]
    total    = np.sum(forcing[:,1:], axis=1)



from setuptools import setup
from setuptools import find_packages
import versioneer

# README #
def readme():
    with open('README.rst') as f:
        return f.read()

AUTHORS = [
    ("Chris Smith", "c.j.smith1@leeds.ac.uk"),
    ("Richard Millar", "richard.millar@ouce.ox.ac.uk"),
    ("Zeb Nicholls", "zebedee.nicholls@climate-energy-college.org"),
    ("Myles Allen", "myles.allen@ouce.ox.ac.uk"),
]

setup(
    name='fair',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description='Python package to perform calculations with the FaIR simple climate model',
    long_description=readme(),
    keywords='simple climate model temperature response carbon cycle emissions forcing',
    url='https://github.com/OMS-NetZero/FAIR',
    author=", ".join([author[0] for author in AUTHORS]),
    author_email=", ".join([author[1] for author in AUTHORS]),
    license='Apache 2.0',
    packages=find_packages(exclude=['tests*','docs*']),
    package_data={'': ['*.csv']},
    python_requires='>=3.6, <4',
    include_package_data=True,
    install_requires=[
        'matplotlib',
        'numpy>=1.14.5',
        'scipy>=0.19.0',
        'pandas'
    ],
    zip_safe=False,
    extras_require={
        'docs': ['sphinx>=1.4', 'nbsphinx'],
        'dev' : ['notebook', 'scmdata<0.6', 'wheel', 'twine'],
        'test': ['pytest>=4.0', 'nbval', 'pytest-cov', 'codecov']
    }
)



from __future__ import division

import numpy as np

def linear(F_CH4, ratio=0.15):
    """Calculates radiative forcing from oxidation of methane to H2O.

    Stratospheric water vapour forcing follows a practically linear
    relationship with the CH4 radiative forcing in MAGICC and AR5.
    """

    F_H2O = ratio * F_CH4
    return F_H2O




from __future__ import division

import numpy as np
from ..constants import cl_atoms, br_atoms, fracrel

def magicc(C_ODS,
           C0, 
           eta1=-1.46030698e-5,
           eta2=2.05401270e-3,
           eta3=1.03143308):

    Cl = np.array(cl_atoms.aslist)
    Br = np.array(br_atoms.aslist)
    FC = np.array(fracrel.aslist)

    EESC = (np.sum(Cl * 1000.*(C_ODS-C0) * FC/FC[0]) +
             45*np.sum(Br * 1000.*(C_ODS-C0) * FC/FC[0])) * FC[0]

    # EESC takes ODS concentrations in ppb, we provide ppt.
    EESC = np.max((EESC,0))

    F = eta1 * (eta2 * EESC) ** eta3
    return F

